import time

from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QSizePolicy
from pyqtgraph.widgets.RawImageWidget import RawImageWidget

from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget


class AttributeImageWidget(AttributeWidget):
    lt = time.perf_counter_ns()

    def _set_input_widget(self):
        self.input_widget = RawImageWidget(scaled=True)
        # self.input_widget = pg.ImageView()
        # self.input_widget = GraphicsView(useOpenGL=True)
        # self.image_item = ImageItem()
        # self.input_widget.addItem(self.image_item)
        if self.attribute[DESCRIPTION] is not None:
            self.input_widget.setToolTip(self.attribute[DESCRIPTION])
        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_value(self, value):
        # self.image_item.setImage(value, autoLevels=False, autoRange=False, scaled=True)
        # self.input_widget.scaleToImage(self.image_item)
        self.input_widget.setImage(value)
        self.input_widget.setMaximumSize(QSize(*value.shape))

    def get_attribute_value(self):
        return None

    def get_widget_value(self):
        return None



import copy
import logging
import random
import time

from numpy import linspace

import kamzik3
from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.macro.common import StepGenerator
from kamzik3.macro.step import StepSetpointNumerical


class Scan(StepGenerator):
    first_step_at = 0
    resolved = False

    def __init__(self, common_id, start_point, end_point, steps_count, step_class=StepSetpointNumerical,
                 step_attributes=None, repeat_count=0, random_deviation=units.Quantity(0, "percent"),
                 wait_after=units.Quantity(0, "s"), retry_count=0, trigger_log=True, as_step=False, bidirectional=False,
                 scanner=None, scanner_attributes=None, return_back=True, on_warning=ABORT, timeout=units.Quantity(0, "s")):
        assert isinstance(start_point, units.Quantity)
        assert isinstance(end_point, units.Quantity)
        assert isinstance(wait_after, units.Quantity)
        assert steps_count > 0
        self.scanner = scanner
        self.scanner_attributes = scanner_attributes
        self.return_back = return_back
        self.step_index = 0
        self.start_point = start_point
        self.end_point = end_point
        self.random_deviation = random_deviation
        if as_step:
            self.steps_count = 1
            self.points_count = 1
        else:
            self.steps_count = steps_count
            self.points_count = steps_count + 1
        self.bidirectional = bidirectional
        self.scan_pass = 0
        self.steps = self.step_generator()
        self.scanning_range = linspace(start_point.m, end_point.m, steps_count + 1)
        self.step_class = step_class
        self.step_attributes = step_attributes
        if self.step_attributes is None:
            self.step_attributes = {}
        self.current_step = None
        self.as_step = as_step
        self.logger = logging.getLogger("Macro.Scan.{}".format(common_id))
        attribute = kamzik3.session.get_device(
            self.step_attributes["device_id"]).get_attribute(self.step_attributes["attribute"])
        self.initial_value_of_attribute = attribute.value()
        super(Scan, self).__init__(common_id, repeat_count, wait_after, retry_count, trigger_log, on_warning, timeout)

    def set_steps_count(self, steps_count):
        """
        Set number of steps for current scan.
        Create step generator within the range of start_point and end_point of size steps_count
        :param steps_count: int
        :return: None
        """
        assert steps_count > 0
        if self.as_step:
            self.steps_count = 1
            self.points_count = 1
        else:
            self.steps_count = steps_count
            self.points_count = steps_count + 1
        self.steps = self.step_generator()
        self.scanning_range = linspace(self.start_point.m, self.end_point.m, steps_count + 1)

    def get_total_steps_count(self):
        """
        Return total number of steps for current Scan
        This number includes all repetitions as well
        :return: int
        """
        repeat_step = (self.step_attributes.get("repeat_count", 0) + 1)
        repeat_scan = (self.repeat_count + 1)
        return self.steps_count * repeat_step * repeat_scan

    def get_total_points_count(self):
        """
        Return total number of points for current Scan
        This number includes all repetitions as well
        :return: int
        """
        repeat_step = (self.step_attributes.get("repeat_count", 0) + 1)
        repeat_scan = (self.repeat_count + 1)
        return self.points_count * repeat_step * repeat_scan

    def step_generator(self):
        """
        Initiate step generator
        :return: Step
        """
        self.first_step_at = time.time()
        scanning_range = self.scanning_range
        if self.bidirectional and self.scan_pass % 2:
            # Reverse scanning range for each odd scan_pass if bidirectional flag is set
            scanning_range = reversed(self.scanning_range)
        for self.step_attributes["common_id"], self.step_attributes["setpoint"] in enumerate(scanning_range):
            # Get through canning range of setpoints and generate step for each one of them
            if self.get_state() == STOPPED:
                # Cancel generator when scan was stopped
                break
            random_offset = 0
            if self.random_deviation.m > 0:
                # If random deviation is set, calculate max size of randomized part of the step
                half_step_size = (abs(self.scanning_range[1] - self.scanning_range[0]) * (
                        self.random_deviation.m / 100)) / 2
                # Generate random offset
                random_offset = random.uniform(-half_step_size, half_step_size)
            self.step_attributes["setpoint"] = units.Quantity(self.step_attributes["setpoint"] + random_offset,
                                                              self.start_point.u)
            # Generate next step based on step attributes
            yield self.step_class(**self.step_attributes)
            # Increase step_index counter
            self.step_index += 1
        # Increase scan pass counter
        self.scan_pass += 1

    def get_reset_step(self):
        """
        Get Step, which initiate Device Attribute into initial Value
        :return: Step
        """
        setpoint_attributes_copy = copy.copy(self.step_attributes)
        setpoint_attributes_copy["common_id"] = "ReturnBack"
        setpoint_attributes_copy["setpoint"] = self.initial_value_of_attribute
        setpoint_attributes_copy["wait_after"] = units.Quantity(0, "s")
        setpoint_attributes_copy["timeout"] = units.Quantity(0, "s")
        try:
            del setpoint_attributes_copy["output"]
        except KeyError:
            pass
        return self.step_class(**setpoint_attributes_copy)

    def return_back_callback(self):
        """
        Whenever Scan was Stopped or aborted, this callback ensures
        that Device attributes get back to initial value.
        :return:
        """
        if self.return_back and self.initial_value_of_attribute is not None:
            # if return_back is set and initial value of attribute is not None
            device = kamzik3.session.get_device(self.step_attributes["device_id"])
            # wait until device is one of the Idle statuses
            if device.wait_for_status(IDLE_DEVICE_STATUSES):
                # get reset step
                step = self.get_reset_step()
                if isinstance(step, list):
                    # we have list of steps to execute
                    for sub_step in step:
                        sub_step.start()
                else:
                    # execute reset step
                    step.start()
            else:
                # device did not get into idle state after timeout of 1s
                self.logger.error(
                    u"Could not return {} to {}, device was not IDLE".format(self.step_attributes["attribute"],
                                                                             self.initial_value_of_attribute))

    def get_output_header(self):
        lines = [
            "Type: {}".format("Scan"),
            "Scanner: {}".format(self.scanner),
            "Device: {}".format(self.step_attributes["device_id"]),
            "Attribute: {}".format(Attribute.list_attribute(self.step_attributes["attribute"])),
            "Start point: {:~}".format(self.start_point),
            "End point: {:~}".format(self.end_point),
            "Steps count: {}".format(self.steps_count),
            "Step size: {:~}".format((self.end_point - self.start_point) / self.steps_count),
            "Points count: {}".format(self.points_count),
            "Bidirectional: {}".format(self.bidirectional),
            "Negative tolerance: {:~}".format(self.step_attributes["negative_tolerance"]),
            "Positive tolerance: {:~}".format(self.step_attributes["positive_tolerance"]),
            "Random deviation: {:~}".format(self.random_deviation),
            "Wait: {:~}".format(self.step_attributes["wait_after"]),
            "Repeat count: {}".format(self.repeat_count),
            "Retry: {}".format(self.retry_count),
            "Return back: {}".format(self.return_back),
            "Trigger log: {}".format(self.step_attributes["trigger_log"]),
        ]
        if self.scanner_attributes is not None:
            for key, value in self.scanner_attributes.items():
                lines.append("{}: {}".format(key, value))
        return lines

    def remove(self):
        if self.current_step is not None:
            self.current_step.remove()
        self.steps = None
        self.scanning_range = None
        self.step_class = None
        self.step_attributes = None
        self.current_step = None
        StepGenerator.remove(self)



import time
from threading import Thread
from time import sleep

import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.dummy.deviceDummy import DeviceDummy
from kamzik3.snippets.snippetsDecorators import expose_method


class DummyMotor(DeviceDummy):

    def __init__(self, channels=3, device_id=None, config=None):
        assert channels > 0
        self.channels = channels
        DeviceDummy.__init__(self, device_id, config)

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_CHANNELS, default_value=self.channels, readonly=True, default_type=np.uint16,
                              max_value=255)


class DummyMotorChannel(DeviceChannel):

    def __init__(self, device, channel, device_id=None, config=None):
        super().__init__(device, channel, device_id, config)
        self.handle_configuration()

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        _finish_configuration()

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)
        self.create_attribute(ATTR_REFERENCED, default_value=True, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_POSITION, default_value=0, default_type=np.float, set_function=self._move_absolute,
                              unit="nm", set_value_when_set_function=False, min_value=-1e9, max_value=1e9)
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_value=-1e9, default_type=np.float, min_value=-1e9,
                              max_value=1e9, unit="nm", set_function=self._set_lower_limit)
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_value=1e9, default_type=np.float, min_value=-1e9,
                              max_value=1e9, unit="nm", set_function=self._set_upper_limit)
        self.create_attribute(ATTR_MAXIMUM_VELOCITY, default_value=1e5, default_type=np.uint64, min_value=10,
                              max_value=1e6, unit="nm/sec", set_function=self._set_maximum_velocity)
        self.create_attribute(ATTR_VELOCITY, default_value=0, default_type=np.float, min_value=-1e4, max_value=1e4,
                              unit="nm/sec")

    def subject_update(self, key, value, subject):
        DeviceChannel.subject_update(self, key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]

            self.set_attribute((ATTR_LATENCY, VALUE), self.device.get_value(ATTR_LATENCY))
            self.set_value(attribute, value)

    def _move_absolute(self, value, callback=None):
        if self.is_status(STATUS_BUSY):
            raise DeviceError("Motor is moving")

        def move_thread(setpoint):
            if value == self.get_value(ATTR_POSITION):
                return

            direction = 1
            if setpoint - self.get_raw_value(ATTR_POSITION) < 0:
                direction = -1

            self.set_raw_value(ATTR_VELOCITY, direction * self.get_raw_value(ATTR_MAXIMUM_VELOCITY))
            self.set_status(STATUS_BUSY)
            counter = 0
            while self.connected and self[ATTR_VELOCITY][VALUE] != 0 and self.is_status(STATUS_BUSY):
                # Velocity in nm / 100ms
                counter += 1
                velocity = int(self.get_raw_value(ATTR_VELOCITY) / 10.)
                new_position = int(self.get_raw_value(ATTR_POSITION) + velocity)
                if direction == 1:
                    if new_position < setpoint:
                        self.set_raw_value(ATTR_POSITION, new_position)
                    else:
                        self.set_raw_value(ATTR_POSITION, setpoint)
                        self.stop()
                else:
                    if new_position > setpoint:
                        self.set_raw_value(ATTR_POSITION, new_position)
                    else:
                        self.set_raw_value(ATTR_POSITION, setpoint)
                        self.stop()
                sleep(0.1)
            self.set_status(STATUS_IDLE)

        Thread(target=move_thread, args=[value]).start()
        if callback is not None:
            callback()

    def _set_upper_limit(self, value, callback=None):
        position_upper_limit = self[ATTR_POSITION].remove_offset_factor(value)
        self.set_raw_value(ATTR_POSITION, position_upper_limit, key=MAX)
        if callback is not None:
            callback()

    def _set_lower_limit(self, value, callback=None):
        position_lower_limit = self[ATTR_POSITION].remove_offset_factor(value)
        self.set_raw_value(ATTR_POSITION, position_lower_limit, key=MIN)
        if callback is not None:
            callback()

    def _set_maximum_velocity(self, velocity_upper_limit, callback=None):
        velocity_lower_limit = -velocity_upper_limit
        self.set_raw_value(ATTR_VELOCITY, velocity_lower_limit, key=MIN)
        self.set_raw_value(ATTR_VELOCITY, velocity_upper_limit, key=MAX)

        if self.get_raw_value(ATTR_VELOCITY) > velocity_upper_limit:
            self.get_raw_value(ATTR_VELOCITY, velocity_upper_limit)
        elif self.get_raw_value(ATTR_VELOCITY) < velocity_lower_limit:
            self.get_raw_value(ATTR_VELOCITY, velocity_lower_limit)
        if callback is not None:
            callback()

    @expose_method()
    def find_reference(self, callback=None):
        if self.is_status(STATUS_BUSY):
            raise DeviceError("Motor is moving")

        self.logger.info(u"Find reference")
        self._set((ATTR_POSITION, VALUE), 0)
        if callback is not None:
            callback()

    @expose_method()
    def stop(self, callback=None):
        self._set([ATTR_VELOCITY, VALUE], 0)
        if callback is not None:
            callback()

    def disconnect(self):
        self.stop()
        super().disconnect()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'customOpenDialogTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(930, 541)
        Dialog.setStyleSheet("QSplitter::handle:horizontal {\n"
"  image: url(:/icons/icons/splitter.png);\n"
"  width:13px;\n"
"  height:13px;\n"
"  background: #f6b442;\n"
"}")
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.splitter = QtWidgets.QSplitter(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setHandleWidth(11)
        self.splitter.setObjectName("splitter")
        self.input_category = QtWidgets.QListView(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.input_category.sizePolicy().hasHeightForWidth())
        self.input_category.setSizePolicy(sizePolicy)
        self.input_category.setMinimumSize(QtCore.QSize(150, 0))
        self.input_category.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.input_category.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.input_category.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.input_category.setResizeMode(QtWidgets.QListView.Adjust)
        self.input_category.setViewMode(QtWidgets.QListView.ListMode)
        self.input_category.setObjectName("input_category")
        self.table_directory_content = QtWidgets.QTableView(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.table_directory_content.sizePolicy().hasHeightForWidth())
        self.table_directory_content.setSizePolicy(sizePolicy)
        self.table_directory_content.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.table_directory_content.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.table_directory_content.setShowGrid(False)
        self.table_directory_content.setSortingEnabled(True)
        self.table_directory_content.setObjectName("table_directory_content")
        self.table_directory_content.horizontalHeader().setDefaultSectionSize(150)
        self.table_directory_content.horizontalHeader().setMinimumSectionSize(150)
        self.table_directory_content.horizontalHeader().setStretchLastSection(False)
        self.table_directory_content.verticalHeader().setVisible(False)
        self.table_directory_content.verticalHeader().setHighlightSections(False)
        self.gridLayout.addWidget(self.splitter, 1, 0, 1, 1)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.input_directory = QtWidgets.QComboBox(Dialog)
        self.input_directory.setObjectName("input_directory")
        self.gridLayout_2.addWidget(self.input_directory, 0, 1, 1, 1)
        self.button_up = QtWidgets.QPushButton(Dialog)
        self.button_up.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_up.sizePolicy().hasHeightForWidth())
        self.button_up.setSizePolicy(sizePolicy)
        self.button_up.setObjectName("button_up")
        self.gridLayout_2.addWidget(self.button_up, 0, 2, 1, 1)
        self.label = QtWidgets.QLabel(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 1, 0, 1, 1)
        self.input_file_name = QtWidgets.QLineEdit(Dialog)
        self.input_file_name.setObjectName("input_file_name")
        self.gridLayout_2.addWidget(self.input_file_name, 1, 1, 1, 1)
        self.button_filter = QtWidgets.QPushButton(Dialog)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/search.png"), QtGui.QIcon.Normal, QtGui.QIcon.On)
        self.button_filter.setIcon(icon)
        self.button_filter.setObjectName("button_filter")
        self.gridLayout_2.addWidget(self.button_filter, 1, 2, 1, 1)
        self.gridLayout.addLayout(self.gridLayout_2, 0, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Open)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 1)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        self.table_directory_content.doubleClicked['QModelIndex'].connect(Dialog.slot_file_selected)
        self.button_up.clicked.connect(Dialog.slot_dir_up)
        self.input_directory.currentIndexChanged['QString'].connect(Dialog.slot_change_path)
        self.button_filter.clicked.connect(Dialog.slot_file_filter)
        self.input_category.clicked['QModelIndex'].connect(Dialog.slot_drive_changed)
        self.input_file_name.returnPressed.connect(self.button_filter.click)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.input_file_name, self.button_filter)
        Dialog.setTabOrder(self.button_filter, self.button_up)
        Dialog.setTabOrder(self.button_up, self.input_directory)
        Dialog.setTabOrder(self.input_directory, self.input_category)
        Dialog.setTabOrder(self.input_category, self.table_directory_content)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.button_up.setText(_translate("Dialog", "^ Up"))
        self.label.setText(_translate("Dialog", "Current directory:"))
        self.label_2.setText(_translate("Dialog", "File name:"))
        self.button_filter.setText(_translate("Dialog", "Filter"))
from kamzik3 import resource_kamzik3_rc



import time

import numpy as np

from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units


class DeviceShutter(Device):

    def __init__(self, device_id=None, config=None):
        super(DeviceShutter, self).__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        super(DeviceShutter, self)._init_attributes()
        self.create_attribute(ATTR_OPENED, default_value=False, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_AUTO_MODE, default_value=False, default_type=np.bool)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def open_shutter(self):
        raise NotImplementedError()

    def close_shutter(self):
        raise NotImplementedError()

    def is_opened(self):
        raise NotImplementedError()


class DeviceShutterAttribute(DeviceShutter):

    def __init__(self, device, attribute, opened_value, closed_value, device_id=None, config=None):
        self.device = device
        self.attribute = Attribute.list_attribute(attribute)
        self.opened_value = opened_value
        self.closed_value = closed_value
        self.configured = False
        super(DeviceShutterAttribute, self).__init__(device_id, config)

    def _init_attributes(self):
        super(DeviceShutterAttribute, self)._init_attributes()
        # opened_value = device_units(self.device, self.attribute, self.opened_value)
        # closed_value = device_units(self.device, self.attribute, self.closed_value)
        # self.create_attribute(ATTR_OPENED_VALUE, default_value=opened_value.m, default_type=np.float64,
        #                       unit="{:~}".format(opened_value.u))
        # self.create_attribute(ATTR_CLOSED_VALUE, default_value=closed_value.m, default_type=np.float64,
        #                       unit="{:~}".format(closed_value.u))
        # self.create_attribute(ATTR_OPENED_AT, default_value=0, readonly=True, default_type=np.float16, min_value=0,
        #                       max_value=100, decimals=2, unit=u"%")
        self.create_attribute(ATTR_OPENED_VALUE, default_value=self.opened_value, default_type=np.float64)
        self.create_attribute(ATTR_CLOSED_VALUE, default_value=self.closed_value, default_type=np.float64)
        self.create_attribute(ATTR_OPENED_AT, default_value=0, readonly=True, default_type=np.float16, min_value=0,
                              max_value=100, decimals=2, unit=u"%")

    def connect(self, *args):
        self.device.attach_attribute_callback(ATTR_STATUS, self.set_status, key_filter=VALUE)

    def set_status(self, status):
        super().set_status(status)
        if status in READY_DEVICE_STATUSES and not self.configured:
            self.handle_configuration()

    def handle_configuration(self):
        self.configured = True
        opened_value = device_units(self.device, self.attribute, self.opened_value)
        closed_value = device_units(self.device, self.attribute, self.closed_value)
        self.set_attribute([ATTR_OPENED_VALUE, UNIT], "{:~}".format(opened_value.u))
        self.set_attribute([ATTR_CLOSED_VALUE, UNIT], "{:~}".format(closed_value.u))
        # self.create_attribute(ATTR_OPENED_VALUE, default_value=opened_value.m, default_type=np.float64,
        #                       unit="{:~}".format(opened_value.u))
        # self.create_attribute(ATTR_CLOSED_VALUE, default_value=closed_value.m, default_type=np.float64,
        #                       unit="{:~}".format(closed_value.u))
        self.device.attach_attribute_callback(self.attribute, self.shutter_value_changed, key_filter=VALUE)

    def shutter_value_changed(self, value):
        self.set_value(ATTR_OPENED, self.is_opened(value))

    @expose_method()
    def open_shutter(self, clear_auto_mode=True):
        self.set_value(ATTR_AUTO_MODE, not clear_auto_mode)
        self.device.set_value(self.attribute, self.get_value(ATTR_OPENED_VALUE))

    @expose_method()
    def close_shutter(self, clear_auto_mode=True):
        self.set_value(ATTR_AUTO_MODE, not clear_auto_mode)
        self.device.set_value(self.attribute, self.get_value(ATTR_CLOSED_VALUE))

    @expose_method()
    def stop(self):
        self.device.stop()

    def is_opened(self, current_value=None):
        if current_value is None:
            return None
        current_value = self.device.get_value(self.attribute)
        tolerance = self.device.get_attribute(self.attribute + [TOLERANCE])
        opened_value = self.get_value(ATTR_OPENED_VALUE)
        try:
            self.set_value(ATTR_OPENED_AT, (float(current_value) / float(opened_value)) * 100)
        except ZeroDivisionError:
            self.set_value(ATTR_OPENED_AT, 100)
        return (opened_value - tolerance[0]) <= current_value <= (opened_value + tolerance[1])

    def close(self):
        self.device.detach_attribute_callback(self.attribute, self.shutter_value_changed)
        self.device.detach_attribute_callback(ATTR_STATUS, self.shutter_status_changed)
        return super(DeviceShutterAttribute, self).close()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'savedAttributesTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1060, 502)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.frame = QtWidgets.QFrame(Form)
        self.frame.setEnabled(True)
        self.frame.setStyleSheet("QFrame {background-color:#d5d5d5}\n"
"QLabel {padding:3px;padding-left:7px;padding-right:7px}")
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(self.frame)
        self.label.setMaximumSize(QtCore.QSize(9999999, 16777215))
        self.label.setFrameShape(QtWidgets.QFrame.Box)
        self.label.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(self.frame)
        self.label_2.setFrameShape(QtWidgets.QFrame.Box)
        self.label_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.label_3 = QtWidgets.QLabel(self.frame)
        self.label_3.setFrameShape(QtWidgets.QFrame.Box)
        self.label_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(self.frame)
        self.label_4.setStyleSheet("")
        self.label_4.setFrameShape(QtWidgets.QFrame.Box)
        self.label_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout.addWidget(self.label_4)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.gridLayout.addWidget(self.frame, 3, 0, 1, 1)
        self.groups_frame = QtWidgets.QFrame(Form)
        self.groups_frame.setStyleSheet("QFrame#groups_frame {background-color:#f6b442}")
        self.groups_frame.setObjectName("groups_frame")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.groups_frame)
        self.horizontalLayout_3.setContentsMargins(6, 2, 6, 2)
        self.horizontalLayout_3.setSpacing(3)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_saved_groups = QtWidgets.QLabel(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_saved_groups.sizePolicy().hasHeightForWidth())
        self.label_saved_groups.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(False)
        font.setWeight(50)
        self.label_saved_groups.setFont(font)
        self.label_saved_groups.setObjectName("label_saved_groups")
        self.horizontalLayout_3.addWidget(self.label_saved_groups)
        self.combo_preset = QtWidgets.QComboBox(self.groups_frame)
        self.combo_preset.setMinimumSize(QtCore.QSize(200, 0))
        self.combo_preset.setObjectName("combo_preset")
        self.horizontalLayout_3.addWidget(self.combo_preset)
        self.button_remove_preset_group = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_remove_preset_group.sizePolicy().hasHeightForWidth())
        self.button_remove_preset_group.setSizePolicy(sizePolicy)
        self.button_remove_preset_group.setMinimumSize(QtCore.QSize(100, 0))
        self.button_remove_preset_group.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(False)
        font.setWeight(50)
        self.button_remove_preset_group.setFont(font)
        self.button_remove_preset_group.setStyleSheet("QPushButton {color:red}")
        self.button_remove_preset_group.setIconSize(QtCore.QSize(0, 0))
        self.button_remove_preset_group.setObjectName("button_remove_preset_group")
        self.horizontalLayout_3.addWidget(self.button_remove_preset_group)
        self.button_add_preset_group = QtWidgets.QPushButton(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_add_preset_group.sizePolicy().hasHeightForWidth())
        self.button_add_preset_group.setSizePolicy(sizePolicy)
        self.button_add_preset_group.setMinimumSize(QtCore.QSize(100, 0))
        self.button_add_preset_group.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(False)
        font.setWeight(50)
        self.button_add_preset_group.setFont(font)
        self.button_add_preset_group.setStyleSheet("QPushButton {color:green}")
        self.button_add_preset_group.setIconSize(QtCore.QSize(0, 0))
        self.button_add_preset_group.setObjectName("button_add_preset_group")
        self.horizontalLayout_3.addWidget(self.button_add_preset_group)
        self.label_saved_groups_2 = QtWidgets.QLabel(self.groups_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_saved_groups_2.sizePolicy().hasHeightForWidth())
        self.label_saved_groups_2.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setPointSize(11)
        font.setBold(False)
        font.setWeight(50)
        self.label_saved_groups_2.setFont(font)
        self.label_saved_groups_2.setObjectName("label_saved_groups_2")
        self.horizontalLayout_3.addWidget(self.label_saved_groups_2)
        self.button_first_page = QtWidgets.QPushButton(self.groups_frame)
        self.button_first_page.setMaximumSize(QtCore.QSize(25, 16777215))
        self.button_first_page.setObjectName("button_first_page")
        self.horizontalLayout_3.addWidget(self.button_first_page)
        self.button_previous_page = QtWidgets.QPushButton(self.groups_frame)
        self.button_previous_page.setMaximumSize(QtCore.QSize(25, 16777215))
        self.button_previous_page.setObjectName("button_previous_page")
        self.horizontalLayout_3.addWidget(self.button_previous_page)
        self.input_page = QtWidgets.QSpinBox(self.groups_frame)
        self.input_page.setMinimum(1)
        self.input_page.setMaximum(10000)
        self.input_page.setObjectName("input_page")
        self.horizontalLayout_3.addWidget(self.input_page)
        self.button_next_page = QtWidgets.QPushButton(self.groups_frame)
        self.button_next_page.setMaximumSize(QtCore.QSize(25, 16777215))
        self.button_next_page.setObjectName("button_next_page")
        self.horizontalLayout_3.addWidget(self.button_next_page)
        self.button_last_page = QtWidgets.QPushButton(self.groups_frame)
        self.button_last_page.setMaximumSize(QtCore.QSize(25, 16777215))
        self.button_last_page.setObjectName("button_last_page")
        self.horizontalLayout_3.addWidget(self.button_last_page)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.checkbox_allow_double_click = QtWidgets.QCheckBox(self.groups_frame)
        self.checkbox_allow_double_click.setObjectName("checkbox_allow_double_click")
        self.horizontalLayout_3.addWidget(self.checkbox_allow_double_click)
        self.gridLayout.addWidget(self.groups_frame, 2, 0, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.button_set_value = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_set_value.sizePolicy().hasHeightForWidth())
        self.button_set_value.setSizePolicy(sizePolicy)
        self.button_set_value.setMinimumSize(QtCore.QSize(35, 35))
        self.button_set_value.setMaximumSize(QtCore.QSize(40, 40))
        self.button_set_value.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/target.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_set_value.setIcon(icon)
        self.button_set_value.setIconSize(QtCore.QSize(25, 25))
        self.button_set_value.setObjectName("button_set_value")
        self.verticalLayout.addWidget(self.button_set_value)
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem2)
        self.button_change_color = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_change_color.sizePolicy().hasHeightForWidth())
        self.button_change_color.setSizePolicy(sizePolicy)
        self.button_change_color.setMinimumSize(QtCore.QSize(35, 35))
        self.button_change_color.setMaximumSize(QtCore.QSize(40, 40))
        font = QtGui.QFont()
        font.setPointSize(11)
        self.button_change_color.setFont(font)
        self.button_change_color.setStyleSheet("QPushButton {color:red}")
        self.button_change_color.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/color_picker.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_change_color.setIcon(icon1)
        self.button_change_color.setIconSize(QtCore.QSize(25, 25))
        self.button_change_color.setObjectName("button_change_color")
        self.verticalLayout.addWidget(self.button_change_color)
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem3)
        self.button_remove = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_remove.sizePolicy().hasHeightForWidth())
        self.button_remove.setSizePolicy(sizePolicy)
        self.button_remove.setMinimumSize(QtCore.QSize(35, 35))
        self.button_remove.setMaximumSize(QtCore.QSize(40, 40))
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.button_remove.setFont(font)
        self.button_remove.setStyleSheet("QPushButton {color:red}")
        self.button_remove.setIconSize(QtCore.QSize(0, 0))
        self.button_remove.setObjectName("button_remove")
        self.verticalLayout.addWidget(self.button_remove)
        spacerItem4 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem4)
        self.button_add = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_add.sizePolicy().hasHeightForWidth())
        self.button_add.setSizePolicy(sizePolicy)
        self.button_add.setMinimumSize(QtCore.QSize(35, 35))
        self.button_add.setMaximumSize(QtCore.QSize(40, 40))
        font = QtGui.QFont()
        font.setPointSize(15)
        font.setBold(True)
        font.setWeight(75)
        self.button_add.setFont(font)
        self.button_add.setStyleSheet("QPushButton {color:green}")
        self.button_add.setIconSize(QtCore.QSize(0, 0))
        self.button_add.setObjectName("button_add")
        self.verticalLayout.addWidget(self.button_add)
        spacerItem5 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.MinimumExpanding)
        self.verticalLayout.addItem(spacerItem5)
        self.gridLayout.addLayout(self.verticalLayout, 1, 2, 1, 1)
        self.table_attributes = QtWidgets.QTableView(Form)
        self.table_attributes.setObjectName("table_attributes")
        self.gridLayout.addWidget(self.table_attributes, 1, 0, 1, 1)
        self.layout_subgroups_select = QtWidgets.QHBoxLayout()
        self.layout_subgroups_select.setObjectName("layout_subgroups_select")
        self.gridLayout.addLayout(self.layout_subgroups_select, 0, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_remove.clicked.connect(Form.remove_rows)
        self.button_add.clicked.connect(Form.add_row)
        self.button_change_color.clicked.connect(Form.select_color)
        self.button_set_value.clicked.connect(Form.set_attribute_value)
        self.button_add_preset_group.clicked.connect(Form.add_preset_group)
        self.button_remove_preset_group.clicked.connect(Form.remove_preset_group)
        self.combo_preset.currentTextChanged['QString'].connect(Form.slot_refresh_group_tables)
        self.input_page.valueChanged['int'].connect(Form.slot_jump_to_page)
        self.button_next_page.clicked.connect(self.input_page.stepUp)
        self.button_previous_page.clicked.connect(self.input_page.stepDown)
        self.button_first_page.clicked.connect(Form.slot_jump_to_first_page)
        self.button_last_page.clicked.connect(Form.slot_jump_to_last_page)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "ALT+A   [ add row ]"))
        self.label_2.setText(_translate("Form", "ALT+D   [ remove row ]"))
        self.label_3.setText(_translate("Form", "ALT+X   [ set value ]"))
        self.label_4.setText(_translate("Form", "ALT+C   [ set row color ]"))
        self.label_saved_groups.setText(_translate("Form", "Group:"))
        self.button_remove_preset_group.setToolTip(_translate("Form", "Remove selected group"))
        self.button_remove_preset_group.setText(_translate("Form", "- Remove group"))
        self.button_add_preset_group.setToolTip(_translate("Form", "Add new save group"))
        self.button_add_preset_group.setText(_translate("Form", "+ Empty group"))
        self.label_saved_groups_2.setText(_translate("Form", "Paging:"))
        self.button_first_page.setToolTip(_translate("Form", "Jump to first page"))
        self.button_first_page.setText(_translate("Form", "<|"))
        self.button_previous_page.setToolTip(_translate("Form", "Jump to previous page"))
        self.button_previous_page.setText(_translate("Form", "<"))
        self.button_next_page.setToolTip(_translate("Form", "Jump to next page"))
        self.button_next_page.setText(_translate("Form", ">"))
        self.button_last_page.setToolTip(_translate("Form", "Jump to last page"))
        self.button_last_page.setText(_translate("Form", "|>"))
        self.checkbox_allow_double_click.setText(_translate("Form", "Allow set on double click"))
        self.button_set_value.setToolTip(_translate("Form", "Set attribute to saved value (ALT+x)"))
        self.button_change_color.setToolTip(_translate("Form", "Change row color (ALT+C)"))
        self.button_remove.setToolTip(_translate("Form", "Remove row (ALT+D)"))
        self.button_remove.setText(_translate("Form", "-"))
        self.button_add.setToolTip(_translate("Form", "Add row (ALT+A)"))
        self.button_add.setText(_translate("Form", "+"))
from kamzik3 import resource_kamzik3_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'deviceGroupTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(367, 212)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(2, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.frame_label_area = QtWidgets.QFrame(Form)
        self.frame_label_area.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_label_area.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_label_area.setLineWidth(0)
        self.frame_label_area.setObjectName("frame_label_area")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame_label_area)
        self.horizontalLayout_2.setContentsMargins(2, 2, 2, 2)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.button_group_name = QtWidgets.QPushButton(self.frame_label_area)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_group_name.sizePolicy().hasHeightForWidth())
        self.button_group_name.setSizePolicy(sizePolicy)
        self.button_group_name.setMaximumSize(QtCore.QSize(16777215, 21))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.button_group_name.setFont(font)
        self.button_group_name.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.button_group_name.setStyleSheet("QPushButton {\n"
"background:transparent;\n"
"}")
        self.button_group_name.setFlat(True)
        self.button_group_name.setObjectName("button_group_name")
        self.horizontalLayout_2.addWidget(self.button_group_name)
        self.button_toggle_attach = QtWidgets.QPushButton(self.frame_label_area)
        self.button_toggle_attach.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.button_toggle_attach.setStyleSheet("QPushButton { background-color: transpare;padding:0;border:0;background-color:transparent }\n"
"QPushButton:pressed { background-color: transparent }")
        self.button_toggle_attach.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/detach.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_toggle_attach.setIcon(icon)
        self.button_toggle_attach.setIconSize(QtCore.QSize(20, 20))
        self.button_toggle_attach.setCheckable(False)
        self.button_toggle_attach.setChecked(False)
        self.button_toggle_attach.setAutoExclusive(True)
        self.button_toggle_attach.setObjectName("button_toggle_attach")
        self.horizontalLayout_2.addWidget(self.button_toggle_attach)
        self.horizontalLayout.addWidget(self.frame_label_area)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)
        self.button_device_group = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_device_group.sizePolicy().hasHeightForWidth())
        self.button_device_group.setSizePolicy(sizePolicy)
        self.button_device_group.setMinimumSize(QtCore.QSize(50, 50))
        self.button_device_group.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        self.button_device_group.setFocusPolicy(QtCore.Qt.NoFocus)
        self.button_device_group.setAutoFillBackground(False)
        self.button_device_group.setStyleSheet("QPushButton#button_device_group {{border:2px solid transparent; border-bottom:0px; border-left:2px solid transparent; border-right:2px solid transparent; background-color:transparent; margin:0px; padding:5px}} :checked {background-color: transparent} :pressed {border:0;background-color: transparent}")
        self.button_device_group.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/deviceExample.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_device_group.setIcon(icon1)
        self.button_device_group.setIconSize(QtCore.QSize(0, 0))
        self.button_device_group.setCheckable(True)
        self.button_device_group.setAutoDefault(False)
        self.button_device_group.setDefault(False)
        self.button_device_group.setFlat(True)
        self.button_device_group.setObjectName("button_device_group")
        self.gridLayout.addWidget(self.button_device_group, 2, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_toggle_attach.clicked.connect(Form.toggle_attach)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.button_group_name.setText(_translate("Form", "Stages"))
        self.button_toggle_attach.setToolTip(_translate("Form", "Detach / attach group"))

from kamzik3 import resource_kamzik3_rc



import json
import logging
from threading import Thread

from kamzik3.devices.deviceClient import DeviceClient

import kamzik3
from kamzik3 import CommandFormatException, DeviceServerError, DeviceError
from kamzik3.constants import *
from kamzik3.snippets.snippetDataAccess import rgetattr
from kamzik3.snippets.snippetsJson import JsonKamzikHook, JsonKamzikEncoder
from kamzik3.snippets.snippetsZmq import Worker


class DeviceWorker(Worker, Thread):

    def __init__(self, id, master_host, master_port, zmq_context=None):
        self.logger = logging.getLogger(u"Console.DeviceWorker@{}".format(id))
        Thread.__init__(self)
        Worker.__init__(self, id, master_host, master_port, zmq_context)
        self.instructions_set += [INSTRUCTION_COMMAND, INSTRUCTION_GET, INSTRUCTION_SET, INSTRUCTION_METHOD,
                                  INSTRUCTION_POLL, INSTRUCTION_INIT]

    def get_payload(self, data):
        try:
            instruction, device_id = data[0], data[1].decode()
            device = kamzik3.session.get_device(device_id)

            # Execute command directly on devices
            if instruction == INSTRUCTION_COMMAND:
                command, with_token = data[2].decode(), int(data[3])
                token = device.command(command, with_token=int(with_token))
                status, response = RESPONSE_OK, str(len(data[3]))

            # Get devices attribute
            elif instruction == INSTRUCTION_GET:
                attribute = json.loads(data[2].decode(), object_hook=JsonKamzikHook)
                response = device._get(attribute)
                response = json.dumps(response, cls=JsonKamzikEncoder, ensure_ascii=True)
                status, token = RESPONSE_OK, 0

            # Set devices attribute
            elif instruction == INSTRUCTION_SET:
                attribute, attribute_value = json.loads(data[2].decode(), object_hook=JsonKamzikHook)
                set_token = device.set_attribute(attribute, attribute_value)
                status, token, response = RESPONSE_OK, set_token, str(len(data[2]))

            # Execute devices method with attributes
            elif instruction == INSTRUCTION_METHOD:
                method, attributes = data[2].decode(), json.loads(data[3].decode(), object_hook=JsonKamzikHook)
                response = rgetattr(device, method)(**attributes)
                response = json.dumps(response, cls=JsonKamzikEncoder, ensure_ascii=True)
                status, token = RESPONSE_OK, 0

            # Poll devices for activity
            elif instruction == INSTRUCTION_POLL:
                status, token, response = RESPONSE_OK, 0, "1"

            # Init devices
            elif instruction == INSTRUCTION_INIT:
                if device is None:
                    raise DeviceServerError(
                        "Device {} is not registered on the server or publisher is not ready".format(device_id))
                else:
                    device_proxy = False
                    if isinstance(device, DeviceClient):
                        device_proxy = (device.host, device.port)
                    device_publisher = (kamzik3.session.publisher.host,
                                        kamzik3.session.publisher.port)
                    response = json.dumps(
                        (device.attributes, device.attributes_sharing_map,
                         device.exposed_methods, device.qualified_name, device_proxy, device_publisher), cls=JsonKamzikEncoder,
                        ensure_ascii=True)
                    status, token = RESPONSE_OK, 0
            # None of above, request not implemented
            else:
                status, token, response = RESPONSE_ERROR, 0, str(len(data[3]))

        except CommandFormatException as e:
            status, token, response = RESPONSE_ERROR, 0, "Command format error"
            self.logger.error("Command format error {} {}".format(str(data), e))
        except DeviceServerError as e:
            status, token, response = RESPONSE_ERROR, 0, "Device server error"
            self.logger.error("Device server error {} {}".format(str(data), e))
        except DeviceError as e:
            status, token, response = RESPONSE_ERROR, 0, "Command error: {}".format(e)
            self.logger.error("Command error {} {}".format(str(data), e))
        except (AttributeError, KeyError) as e:
            status, token, response = RESPONSE_ERROR, 0, "Attribute error: {}".format(e)
            self.logger.error("Attribute error {} {}".format(str(data), e))
        except Exception as e:
            status, token, response = RESPONSE_ERROR, 0, "Unknown error: {}".format(e)
            self.logger.error("Unknown error {} {}".format(str(data), e))
        finally:
            return [status.encode(), str(token).encode(), MSG_JSON, response.encode()]



import sys
import psutil

if not psutil.WINDOWS:
    raise Exception("Not supported OS")

from time import perf_counter_ns, sleep
from ctypes import *
from ctypes.wintypes import DWORD
from ctypes.wintypes import UINT
import contextlib
import ctypes
from ctypes import wintypes
from kamzik3.snippets.snippetsInterfaces import CallbackTimerInterface

timeproc = WINFUNCTYPE(None, c_uint, c_uint, DWORD, DWORD, DWORD)
timeSetEvent = windll.winmm.timeSetEvent
timeKillEvent = windll.winmm.timeKillEvent


class CallbackTimerWMT(CallbackTimerInterface):

    def __init__(self, timeout, callback, with_correction=True):
        self.resolution = UINT(0)
        self.tick_function = callback
        self.periodic = True
        self.id = None
        self.calbckfn = timeproc(self.callback)
        CallbackTimerInterface.__init__(self, callback, timeout, with_correction)

    def set_timeout(self, timeout):
        self.interval = UINT(int(timeout))

    def start(self, instant=False):
        if instant:
            self.tick_function()

        self.id = timeSetEvent(self.interval, self.resolution, self.calbckfn, c_ulong(0), c_uint(self.periodic))
        CallbackTimerInterface.start(self)

    def run(self):
        while not self.stopped.wait(0.1):
            pass
        timeKillEvent(self.id)

    def callback(self, uID, uMsg, dwUser, dw1, dw2):
        self.tick_function()

    def stop(self):
        self.stopped.set()


winmm = ctypes.WinDLL('winmm')


class TIMECAPS(ctypes.Structure):
    _fields_ = (('wPeriodMin', wintypes.UINT),
                ('wPeriodMax', wintypes.UINT))


def _check_time_err(err, func, args):
    if err:
        raise WindowsError('%s error %d' % (func.__name__, err))
    return args


winmm.timeGetDevCaps.errcheck = _check_time_err
winmm.timeBeginPeriod.errcheck = _check_time_err
winmm.timeEndPeriod.errcheck = _check_time_err


@contextlib.contextmanager
def timer_resolution(msecs=0):
    caps = TIMECAPS()
    winmm.timeGetDevCaps(ctypes.byref(caps), ctypes.sizeof(caps))
    msecs = min(max(msecs, caps.wPeriodMin), caps.wPeriodMax)
    winmm.timeBeginPeriod(msecs)
    yield
    winmm.timeEndPeriod(msecs)


def set_timer_resolution(msecs=0):
    caps = TIMECAPS()
    winmm.timeGetDevCaps(ctypes.byref(caps), ctypes.sizeof(caps))
    msecs = min(max(msecs, caps.wPeriodMin), caps.wPeriodMax)
    winmm.timeBeginPeriod(msecs)


class CallbackTimerWOC(CallbackTimerInterface):

    def run(self):
        if self.with_correction:
            time_reference = perf_counter_ns()
            with timer_resolution(0):
                while not self.stopped.is_set():
                    try:
                        sleep(self.timeout - (perf_counter_ns() - time_reference) * 1e-9)
                    except ValueError:
                        pass
                    self.callback()
                    time_reference += self.timeout * 1e9
        else:
            with timer_resolution(0):
                while not self.stopped.is_set():
                    sleep(self.timeout)
                    self.callback()


import re
import time

import bidict
import numpy as np
from kamzik3.snippets.snippetsDecorators import expose_method

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.deviceSocket import DeviceSocket

VAL_NEGATIVE = u"Negative"
VAL_POSITIVE = u"Positive"
VAL_POWER_HOLD = u"Power hold"
VAL_POWER_RELEASED = u"Power released"
VAL_OFF = "Off"
VAL_UP_AND_DOWN = "Up and down"
VAL_UP_ONLY = "Up only"
VAL_UP_ONLY_SPECIAL = "Up only - special"
VAL_UP_ONLY_SPECIAL_2 = "Up only - special 2"


class DeviceProbusV(DeviceSocket):
    terminator = b"\n"
    push_commands_max = 1
    polarity_modes = bidict.frozenbidict({
        VAL_POSITIVE: 0,
        VAL_NEGATIVE: 1,
    })
    ramping_modes = bidict.frozenbidict({
        VAL_OFF: 0,
        VAL_UP_AND_DOWN: 1,
        VAL_UP_ONLY: 2,
        VAL_UP_ONLY_SPECIAL: 3,
        VAL_UP_ONLY_SPECIAL_2: 4,
    })
    error_check_pattern = re.compile("E(\d+)")
    error_description = {
        "E0": "No Error",
        "E1": "No data available",
        "E2": "Unknown register type",
        "E4": "Invalid argument",
        "E5": "Argument out of range",
        "E6": "Register is read only",
        "E7": "Receive overflow",
        "E8": "EEPROM is write protected",
        "E9": "Address error",
        "E10": "Unknown SCPI command",
        "E11": "Not allowed Trigger-on-Talk",
        "E12": "Invalid argument in ~Tn command",
        "E13": "Invalid N-value",
        "E14": "Register is write only",
        "E15": "String too long",
        "E16": "Wrong checksum"
    }

    def __init__(self, host, port=2101, udp_host=None, udp_port=None, device_id=None, config=None):
        DeviceSocket.__init__(self, host, port, device_id, config)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)

        self.create_attribute(ATTR_VOLTAGE_SETPOINT, default_type=np.float16, min_value=0, max_value=2000, decimals=3,
                              unit="V", set_function=self.set_voltage)
        self.create_attribute(ATTR_CURRENT_SETPOINT, default_type=np.float16, min_value=0, max_value=1.5, decimals=3,
                              unit="A", set_function=self.set_current)
        self.create_attribute(ATTR_VOLTAGE_READOUT, readonly=True, default_type=np.float16, min_value=0, decimals=3,
                              unit="V")
        self.create_attribute(ATTR_CURRENT_READOUT, readonly=True, default_type=np.float16, min_value=0, decimals=6,
                              unit="A")
        self.create_attribute(ATTR_OUTPUT_RELEASED, default_value=None,
                              default_type=bool, set_function=self.set_release_output)
        self.create_attribute(ATTR_DIGITAL_PROGRAMMING_ACTIVE, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_ANALOGUE_PROGRAMMING_ACTIVE, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_CONTROL_VOLTAGE, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_CONTROL_CURRENT, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_CALIBRATION_MODE, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_POLARITY_REVERSAL, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_CURRENT_ON_TARGET, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_VOLTAGE_ON_TARGET, readonly=True, default_value=None, default_type=bool)
        self.create_attribute(ATTR_RAMP_VOLTAGE_MODE, default_value=None, default_type=self.ramping_modes.keys(),
                              set_function=self.set_voltage_ramping_mode)
        self.create_attribute(ATTR_RAMP_CURRENT_MODE, default_value=None, default_type=self.ramping_modes.keys(),
                              set_function=self.set_current_ramping_mode)
        self.create_attribute(ATTR_OUTPUT_POLARITY, default_value=None, default_type=self.polarity_modes.keys(),
                              set_function=self.set_polarity)
        self.create_attribute(ATTR_VOLTAGE_RAMPING_RATE, default_type=np.float16, min_value=0, max_value=10000,
                              decimals=3,
                              unit="V/s", set_function=self.set_voltage_ramping_rate)
        self.create_attribute(ATTR_CURRENT_RAMPING_RATE, default_type=np.float16, min_value=0, max_value=10000,
                              decimals=3,
                              unit="A/s", set_function=self.set_current_ramping_rate)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.get_digital_programming_output_bit()
        self.get_analog_programming_active_bit()
        self.get_release_output_bit()
        self.get_control_voltage_bit()
        self.get_control_current_bit()
        self.get_polarity_reversal_bit()
        self.get_output_polarity_bit()
        self.get_voltage_ramping_rate()
        self.get_current_ramping_rate()
        self.get_voltage_ramping_mode()
        self.get_current_ramping_mode()
        # Set output measurement for V and A to normal precision
        self.command(">S0H 0\r\n")
        self.command(">S1H 0\r\n", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DeviceSocket.handle_readout(self, readout_buffer)
        # Strip white characters from output
        error_code = self.error_check_pattern.match(output)
        if error_code is not None:
            if (int(error_code.group(1))) != 0:
                # We have error code
                self.handle_command_error(command, self.error_description.get(output, "Unknown error"))
                return
        else:
            (request, value) = output.split(":")
            if command == ">S0?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_SETPOINT, float(value))
            elif command == ">S1?\r\n":
                self.set_raw_value(ATTR_CURRENT_SETPOINT, float(value) * 1000)
            elif command == ">BON?\r\n":
                self.set_raw_value(ATTR_OUTPUT_RELEASED, bool(int(value)))
            elif command == ">KS?\r\n":
                self.set_raw_value(ATTR_CONTROL_CURRENT, bool(int(value[0])))
                self.set_raw_value(ATTR_CONTROL_VOLTAGE, bool(int(value[1])))
                self.set_raw_value(ATTR_OUTPUT_RELEASED, bool(int(value[2])))
                self.set_raw_value(ATTR_CALIBRATION_MODE, bool(int(value[5])))
                self.set_status(STATUS_BUSY if bool(int(value[2])) else STATUS_IDLE)
            elif command == ">DSD?\r\n":
                self.set_raw_value(ATTR_DIGITAL_PROGRAMMING_ACTIVE, bool(int(value)))
            elif command == ">DSA?\r\n":
                self.set_raw_value(ATTR_ANALOGUE_PROGRAMMING_ACTIVE, bool(int(value)))
            elif command == ">DVR?\r\n":
                self.set_raw_value(ATTR_CONTROL_VOLTAGE, bool(int(value)))
            elif command == ">DIR?\r\n":
                self.set_raw_value(ATTR_CONTROL_CURRENT, bool(int(value)))
            elif command == ">DCAL?\r\n":
                self.set_raw_value(ATTR_CALIBRATION_MODE, bool(int(value)))
            elif command == ">BX?\r\n":
                self.set_raw_value(ATTR_POLARITY_REVERSAL, bool(int(value)))
            elif command == ">DX?\r\n":
                self.set_raw_value(ATTR_OUTPUT_POLARITY, self.polarity_modes.inv.get(int(value)))
            elif command == ">S0R?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_RAMPING_RATE, float(value))
            elif command == ">S1R?\r\n":
                self.set_raw_value(ATTR_CURRENT_RAMPING_RATE, float(value))
            elif command == ">M0?\r\n":
                value = float(value)
                if self.get_value(ATTR_POLARITY_REVERSAL):
                    value *= -1
                self.set_raw_value(ATTR_VOLTAGE_READOUT, value)
            elif command == ">M1?\r\n":
                self.set_raw_value(ATTR_CURRENT_READOUT, float(value) * 1000)
            elif command == ">S0S?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_ON_TARGET, bool(int(value)))
            elif command == ">S1S?\r\n":
                self.set_raw_value(ATTR_CURRENT_ON_TARGET, bool(int(value)))
            elif command == ">S0?\r\n":
                self.set_raw_value(ATTR_VOLTAGE_SETPOINT, float(value))
            elif command == ">S0B?\r\n":
                self.set_raw_value(ATTR_RAMP_VOLTAGE_MODE, self.ramping_modes.inv.get(int(value)))
            elif command == ">S1B?\r\n":
                self.set_raw_value(ATTR_RAMP_CURRENT_MODE, self.ramping_modes.inv.get(int(value)))
        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)

        return command, output, callback, token

    def start_polling(self):
        DeviceSocket.start_polling(self)
        self.poll_command(">S0?\r\n", 1000)
        self.poll_command(">S1?\r\n", 1000)
        self.poll_command(">M0?\r\n", 500)
        self.poll_command(">M1?\r\n", 500)
        self.poll_command(">KS?\r\n", 300)
        self.poll_command(">S0S?\r\n", 1000)
        self.poll_command(">S1S?\r\n", 1000)

    def get_voltage_ramping_mode(self):
        self.command(f">S0B?\r\n")

    def get_current_ramping_mode(self):
        self.command(f">S1B?\r\n")

    def get_voltage(self):
        self.command(">S0?\r\n")

    def get_current(self):
        self.command(">S1?\r\n")

    def get_release_output_bit(self):
        self.command(">BON?\r\n")

    def get_polarity_reversal_bit(self):
        self.command(">BX?\r\n")

    def get_control_voltage_bit(self):
        self.command(">DVR?\r\n")

    def get_control_current_bit(self):
        self.command(">DIR?\r\n")

    def get_output_polarity_bit(self):
        self.command(">DX?\r\n")

    def get_digital_programming_output_bit(self):
        self.command(">DSD?\r\n")

    def get_analog_programming_active_bit(self):
        self.command(">DSA?\r\n")

    def get_voltage_readout(self):
        self.command(">M0?\r\n")

    def get_current_readout(self):
        self.command(">M1?\r\n")

    def get_offset(self):
        self.command(">CM0O?\r\n")
        self.command(">CM1O?\r\n")

    def get_gain(self):
        self.command(">CM0GP?\r\n")
        self.command(">CM0GN?\r\n")

    def get_voltage_ramping_rate(self):
        self.command(">S0R?\r\n")

    def get_current_ramping_rate(self):
        self.command(">S1R?\r\n")

    def get_state(self):
        self.command(">KS?\r\n")

    def set_voltage(self, value):
        self.command(f">S0 {value}\r\n")

    def set_current(self, value):
        value = float(value) * 1e-3
        self.command(f">S1 {value:.5e}\r\n")

    def set_release_output(self, value):
        self.command(f">BON {int(value)}\r\n")

    def set_polarity(self, value):
        self.command(f">BX {self.polarity_modes.get(value)}\r\n")

    def set_voltage_ramping_rate(self, value):
        self.command(f">S0R {value}\r\n")

    def set_current_ramping_rate(self, value):
        self.command(f">S1R {value}\r\n")

    def set_voltage_ramping_mode(self, value):
        self.command(f">S0B {self.ramping_modes.get(value)}\r\n")

    def set_current_ramping_mode(self, value):
        self.command(f">S1B {self.ramping_modes.get(value)}\r\n")

    @expose_method()
    def release_output(self):
        self.set_release_output(True)

    @expose_method()
    def stop(self):
        self.set_release_output(False)


import json
import logging
from threading import Thread

import numpy as np

from kamzik3.constants import MSG_JSON, MSG_ARRAY, MSG_BYTES
from kamzik3.snippets.snippetsJson import JsonKamzikEncoder
from kamzik3.snippets.snippetsZmq import Publisher


class DevicePublisher(Publisher, Thread):

    def __init__(self, host, port, zmq_context=None):
        self.logger = logging.getLogger(u"Console.DeviceServer@{}:{}".format(host, port))
        Thread.__init__(self)
        Publisher.__init__(self, host, port, zmq_context)

    def _publish_message(self, header, message):
        try:
            if isinstance(message, np.ndarray):
                self.socket_publisher.send_multipart(
                    [header.encode(), MSG_ARRAY, str(message.dtype.name).encode(), json.dumps(message.shape).encode(),
                     message], copy=False)
            else:
                json_message = json.dumps(message, cls=JsonKamzikEncoder, ensure_ascii=True)
                self.socket_publisher.send_multipart([header.encode(), MSG_JSON, json_message.encode()])
        except Exception as e:
            self.logger.error("{}, {}, {}".format(e, header, message))



import time

import numpy as np
import serial

import kamzik3
from kamzik3 import units, WriteException
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket


class DeviceThermoCube(DeviceSocket):
    """
    The ThermoCube 200/300/400 can store 8 bytes of transmission and can only handle up to 3
    commands per second.
    """
    terminator = b"\r\n"
    push_commands_max = 1

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_FAULT_BYTE, readonly=True, min_value=0, max_value=8, default_type=np.uint8)
        self.create_attribute(ATTR_TEMPERATURE, readonly=True, min_value=-100, max_value=100, unit="degC",
                              default_type=np.float, decimals=3)
        self.create_attribute(ATTR_SETPOINT, min_value=-100, max_value=100, unit="degC", default_type=np.float,
                              decimals=3, set_function=self.set_setpoint)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command(b'\xC8')
        self.command(b'\xC9')
        self.command(b'\xC1', callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DeviceSocket.handle_readout(self, readout_buffer)
        if command == b'\xC9':
            temperature = self.hex_to_temperature(self.parse_temperature_readout(output))
            self.set_raw_value(ATTR_TEMPERATURE, temperature.m)
        elif command == b'\xC1':
            setpoint = self.hex_to_temperature(self.parse_temperature_readout(output))
            self.set_raw_value(ATTR_SETPOINT, setpoint.m)
        elif command == b'\xC8':
            errorByte = int(output)
            self.set_raw_value(ATTR_FAULT_BYTE, errorByte)
            if errorByte == 0:
                self.set_status(STATUS_IDLE)
            else:
                self.set_status(STATUS_ERROR)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)

    def start_polling(self):
        DeviceSocket.start_polling(self)
        # Poll current temperature
        self.poll_command(b'\xC9', 500)
        # Poll current setpoint
        self.poll_command(b'\xC1', 5000)
        # Poll current setpoint
        self.poll_command(b'\xC8', 500)

    @staticmethod
    def parse_temperature_readout(rawDeviceReadout):
        low_byte, high_byte = rawDeviceReadout.split(",")
        return "0x%x%x" % (int(high_byte), int(low_byte))

    @staticmethod
    def temperature_to_hex(temperature):
        temperature_in_kelvin = units.Quantity(temperature, "celsius").to("fahrenheit")
        return "{:04X}".format(int(temperature_in_kelvin.m * 10))

    @staticmethod
    def hex_to_temperature(hexTemperature):
        temperature_in_fahrenheit = int(hexTemperature, 16) / 10
        return units.Quantity(temperature_in_fahrenheit, "fahrenheit").to("celsius")

    def get_temperature(self):
        self.command(b'\xC9')

    def get_setpoint(self):
        self.command(b'\xC1')

    def get_error(self):
        self.command(b'\xC8')

    def set_setpoint(self, temperature, callback=None):
        """
        Set temperature setpoint.
        We need to send three commands.
        E1, Low data byte, High data byte.
        :param temperature:
        :param callback:
        :return:
        """
        temperature_in_hex = self.temperature_to_hex(temperature)
        self.command(b'\xE1', returning=False)
        self.command(bytes.fromhex(temperature_in_hex[2:]), returning=False)
        self.command(bytes.fromhex(temperature_in_hex[:2]), returning=False, with_token=True)

    def send_command(self, commands):
        try:
            if self.connected:
                command = commands.pop(0)
                if command[3]:  # If command is returning
                    self.commands_buffer.append((command, time.time()))
                else:  # Command is not returning, simulate immediate execution
                    self.response_timestamp = time.time()
                self.push(command[0])
            else:
                self.handle_response_error(u"Device is not connected")
        except IndexError:
            self.handle_connection_error(u"Device {} buffer error".format(self.device_id))
        except (WriteException, serial.SerialException):
            self.handle_response_error(u"Device {} writing error".format(self.device_id))
        finally:
            return commands


class DeviceThermoCubePort(DevicePort, DeviceThermoCube):
    terminator = None
    push_commands_max = 1

    def __init__(self, port, baud_rate=9600, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        DevicePort.__init__(self, port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b'\xC8')
        time.sleep(0.3)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            readout = ord(self.read_all())
            if readout >= 0:
                return True
        return False

    def collect_incoming_data(self, data):
        # Get all bytes as separate string
        for c in data:
            self.buffer.append(str(c))

        cmd_query = self.commands_buffer[0][0][0]
        # Commands that queries temperature consist of two bytes
        if cmd_query in (b'\xC9', b'\xC1') and len(self.buffer) != 2:
            return
        # Create final string by joining buffer content with comma
        self.buffer = [",".join(self.buffer)]
        # Invoke found_terminator even though there is now terminator
        # We just want to continue in normal controller workflow and process command
        self.found_terminator()



def expose_method(parameters=None):
    """
    Use this decorator for any devices method.
    Decorated function will be exposed to the client.
    Example:
        @expose_method({"position": "Relative position"})
        def move_relative(self, position):
            ...
    :param parameters: dict(Exposed param1: Description, Exposed param2: Description, ...)
    :return: decorated function
    """

    def decorator(fun):
        setattr(fun, "exposed_parameters", parameters)
        # fun.exposed_parameters = parameters
        return fun

    return decorator


from PyQt5.QtWidgets import QWidget

from kamzik3.gui.deviceDebugWidget import DeviceDebugWidget
from kamzik3.snippets.snippetsWidgets import show_centralized_top


class StackableWidget(QWidget):
    debug_widget = None

    def enterEvent(self, event):
        self.setStyleSheet("QWidget#widget_holder {background:#f1f1f1}")
        if self.model_image is not None:
            self.sigShowModel.emit(self.model_image)
        QWidget.enterEvent(self, event)

    def leaveEvent(self, event):
        self.setStyleSheet("QWidget#widget_holder {background:transparent}")
        if self.model_image is not None:
            self.sigHideModel.emit(self.model_image)
        QWidget.leaveEvent(self, event)

    def show_debug_widget(self):
        if self.debug_widget is None:
            self.debug_widget = DeviceDebugWidget(self.device)

        show_centralized_top(self.debug_widget)

        def close_event(event):
            self.debug_widget.close()
            self.debug_widget.setParent(None)
            self.debug_widget.deleteLater()
            self.debug_widget = None
            event.accept()

        self.debug_widget.closeEvent = close_event



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'macroProgressTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.13.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(913, 462)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_status = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_status.sizePolicy().hasHeightForWidth())
        self.label_status.setSizePolicy(sizePolicy)
        self.label_status.setMinimumSize(QtCore.QSize(100, 0))
        self.label_status.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        font.setKerning(True)
        self.label_status.setFont(font)
        self.label_status.setStyleSheet("QLabel {background:silver}")
        self.label_status.setAlignment(QtCore.Qt.AlignCenter)
        self.label_status.setObjectName("label_status")
        self.horizontalLayout_2.addWidget(self.label_status)
        self.progress_bar = QtWidgets.QProgressBar(Form)
        self.progress_bar.setProperty("value", 0)
        self.progress_bar.setObjectName("progress_bar")
        self.horizontalLayout_2.addWidget(self.progress_bar)
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.label_macro_step_progress = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_macro_step_progress.sizePolicy().hasHeightForWidth())
        self.label_macro_step_progress.setSizePolicy(sizePolicy)
        self.label_macro_step_progress.setMinimumSize(QtCore.QSize(70, 0))
        self.label_macro_step_progress.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.label_macro_step_progress.setStyleSheet("QLabel {background:#fff}")
        self.label_macro_step_progress.setAlignment(QtCore.Qt.AlignCenter)
        self.label_macro_step_progress.setObjectName("label_macro_step_progress")
        self.horizontalLayout_2.addWidget(self.label_macro_step_progress)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.label_running_time = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_running_time.sizePolicy().hasHeightForWidth())
        self.label_running_time.setSizePolicy(sizePolicy)
        self.label_running_time.setMinimumSize(QtCore.QSize(70, 0))
        self.label_running_time.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.label_running_time.setStyleSheet("QLabel {background:#fff}")
        self.label_running_time.setAlignment(QtCore.Qt.AlignCenter)
        self.label_running_time.setObjectName("label_running_time")
        self.horizontalLayout_2.addWidget(self.label_running_time)
        self.label_3 = QtWidgets.QLabel(Form)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_2.addWidget(self.label_3)
        self.label_remaining_time = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_remaining_time.sizePolicy().hasHeightForWidth())
        self.label_remaining_time.setSizePolicy(sizePolicy)
        self.label_remaining_time.setMinimumSize(QtCore.QSize(70, 0))
        self.label_remaining_time.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.label_remaining_time.setStyleSheet("QLabel {background:#fff}")
        self.label_remaining_time.setAlignment(QtCore.Qt.AlignCenter)
        self.label_remaining_time.setObjectName("label_remaining_time")
        self.horizontalLayout_2.addWidget(self.label_remaining_time)
        spacerItem = QtWidgets.QSpacerItem(0, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.label_4 = QtWidgets.QLabel(Form)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_2.addWidget(self.label_4)
        self.label_estimated_finish = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_estimated_finish.sizePolicy().hasHeightForWidth())
        self.label_estimated_finish.setSizePolicy(sizePolicy)
        self.label_estimated_finish.setMinimumSize(QtCore.QSize(70, 0))
        self.label_estimated_finish.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.label_estimated_finish.setStyleSheet("QLabel {background:#fff}")
        self.label_estimated_finish.setAlignment(QtCore.Qt.AlignCenter)
        self.label_estimated_finish.setObjectName("label_estimated_finish")
        self.horizontalLayout_2.addWidget(self.label_estimated_finish)
        self.button_stop = QtWidgets.QPushButton(Form)
        self.button_stop.setStyleSheet("QPushButton { background-color: #ff4e00 }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/stop.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_stop.setIcon(icon)
        self.button_stop.setIconSize(QtCore.QSize(16, 16))
        self.button_stop.setObjectName("button_stop")
        self.horizontalLayout_2.addWidget(self.button_stop)
        self.button_start = QtWidgets.QPushButton(Form)
        self.button_start.setStyleSheet("QPushButton { background-color: #68d532 }\n"
"QPushButton:pressed { background-color: #28a900 }\n"
"QPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\n"
"QPushButton:disabled{ background:silver;color:grey }")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/play.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_start.setIcon(icon1)
        self.button_start.setIconSize(QtCore.QSize(16, 16))
        self.button_start.setObjectName("button_start")
        self.horizontalLayout_2.addWidget(self.button_start)
        self.gridLayout.addLayout(self.horizontalLayout_2, 2, 0, 1, 1)
        self.scroll_area_plot = QtWidgets.QScrollArea(Form)
        self.scroll_area_plot.setWidgetResizable(True)
        self.scroll_area_plot.setObjectName("scroll_area_plot")
        self.plot_widget = QtWidgets.QWidget()
        self.plot_widget.setGeometry(QtCore.QRect(0, 0, 911, 426))
        self.plot_widget.setObjectName("plot_widget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.plot_widget)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.scroll_area_plot.setWidget(self.plot_widget)
        self.gridLayout.addWidget(self.scroll_area_plot, 1, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_stop.clicked.connect(Form.stop)
        self.button_start.clicked.connect(Form.start)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_status.setText(_translate("Form", "Disconnected"))
        self.progress_bar.setToolTip(_translate("Form", "Current scan progress"))
        self.label.setText(_translate("Form", "Points:"))
        self.label_macro_step_progress.setToolTip(_translate("Form", "Current scan step"))
        self.label_macro_step_progress.setText(_translate("Form", "0 / 0"))
        self.label_2.setText(_translate("Form", "Running:"))
        self.label_running_time.setToolTip(_translate("Form", "Current scan running time"))
        self.label_running_time.setText(_translate("Form", "00:00:00"))
        self.label_3.setText(_translate("Form", "Remaining:"))
        self.label_remaining_time.setToolTip(_translate("Form", "Current scan remaining time"))
        self.label_remaining_time.setText(_translate("Form", "00:00:00"))
        self.label_4.setText(_translate("Form", "Estimated end:"))
        self.label_estimated_finish.setToolTip(_translate("Form", "Current scan estimated finish"))
        self.label_estimated_finish.setText(_translate("Form", "00:00:00"))
        self.button_stop.setToolTip(_translate("Form", "Stop current scan"))
        self.button_stop.setText(_translate("Form", "Stop"))
        self.button_start.setToolTip(_translate("Form", "Resume current scan"))
        self.button_start.setText(_translate("Form", "Start"))
from kamzik3 import resource_kamzik3_rc



import json
import pickle
import re
from threading import Event, Thread

import numpy as np
import zmq

from kamzik3.constants import *
from kamzik3.snippets.snippetsJson import JsonKamzikHook


class Listener:

    def __init__(self, host, port, topics=None, token_filter=None):
        """
        Simple listener for Kamzik3 broadcast
        :param host: host of broadcaster
        :param port: port of broadcaster
        :param topics: list of topics subscribe
        :param token_filter: regexp for token filtering
        """
        self.host = host
        self.port = port
        self.topics = topics
        self.stopped = Event()
        self.token_filter = re.compile(".") if token_filter is None else token_filter
        self.socket = zmq.Context.instance().socket(zmq.SUB)
        self.connect()

    def connect(self):
        """
        Connect to broadcaster
        """
        # Set receive timeout to 100 ms
        self.socket.setsockopt(zmq.RCVTIMEO, 100)
        # Don't let socket linger after close
        # This is VERY IMPORTANT not to set to anything but 0
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.connect(f"tcp://{self.host}:{self.port}")
        # Subscribe to topics of interest
        for topic in self.topics:
            self.socket.setsockopt_string(zmq.SUBSCRIBE, topic)

    def listen(self, blocking=True):
        """
        Start listening to broadcast in blocking or non blocking way
        :param blocking: bool flag
        """
        if blocking:
            self._listen()
        else:
            Thread(target=self._listen).start()

    def _listen(self):
        """
        Collect and handle broadcast message until stopped flag is set.
        Kamzik3 is broadcasting many attributes and it's values.
        Format is:
            DeviceId.(type of token).(token header)
        Usually You are interested in Attribute which it token type: A!
        Token header for attribute is in format (Attribute name).(Value | Type | Min | Max | ...)
        For Your application You need only Value.
        Some attributes can have attribute group.
        In that case Attribute name has two levels (Attribute Group.Attribute Name)
        """
        while not self.stopped.isSet():
            if self.socket.poll(timeout=100):
                # Try to read any message from publisher
                reply = self.socket.recv_multipart()
                token, stype = reply[:2]
                # Deserialize data based on type received
                if stype == MSG_JSON:
                    data = json.loads(reply[2].decode(), object_hook=JsonKamzikHook)
                elif stype == MSG_ARRAY:
                    dtype, shape = reply[2:4]
                    reply[4] = np.frombuffer(reply[4], dtype=dtype.decode())
                    data = np.reshape(reply[4], json.loads(shape.decode()))
                elif stype == MSG_PICKLE:
                    data = pickle.loads(reply[2])
                self.handle_readout(token.decode(), data)

    def handle_readout(self, token, data):
        """
        Override this function to do some action upon received message
        """
        if self.token_filter.search(token):
            print(token, data)

    def close(self):
        """
        Set closing flag and stop listener
        """
        self.stopped.set()


if __name__ == "__main__":
    token_filter = re.compile(".Point index\.Value|.Header\.Value|Scan_[0-9]+.*Status")
    kamzik_listener = Listener("127.0.0.1", 50002, ["Scan"], token_filter)
    kamzik_listener.listen(blocking=False)



import collections
import math
from time import perf_counter_ns

import pint
from numpy import dtype, float16, float32, float64, uint16, uint32, uint64, int16, int32, int64, uint8, int8, ndarray

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.constants import SAVE_CHANGE
from kamzik3.snippets.snippetDataAccess import is_equal


class SetFunction(object):

    def __init__(self, attribute, callback=None):
        self.attribute = attribute
        self.callback = callback

    def __enter__(self):
        self.attribute.with_set_function = True
        self.attribute.set_callback = self.callback
        return self.attribute

    def __exit__(self, value_type, value, traceback):
        self.attribute.with_set_function = False
        self.attribute.set_callback = None


class Attribute(dict):
    """
    Class definition of Attribute.
    """
    # Holds main attributes, which each dict has to contain in order to be convertible into Attribute
    attribute_keys = {VALUE, TYPE, UNIT, READONLY, DESCRIPTION}
    integer_types = (int8, int16, int32, int64, uint8, uint16, uint32, uint64)
    float_types = (float16, float32, float64)
    unsigned_types = (uint8, uint16, uint32, uint64)
    # Helper flag for SetFunction decorator
    with_set_function = False
    # Holds last time, that Attribute value was streamed to clients
    last_stream_time = last_update_time = perf_counter_ns()
    # Holds diff between last_stream_time and current time
    last_update_time_diff = 0
    numerical = False
    # Token identifier associate with setting of new value
    set_token = 0
    set_callback = None
    # Minimum time in between of value change stream.
    # Every Attribute change between 0 and min_stream_timeout will be ignored.
    # Set this attribute to zero for Attribute that You wont to stream every change.
    # Default value is in nanoseconds (100ms)
    min_stream_timeout = 100 * 1e6

    def __init__(self, attribute_id=None, default_value=None, min_value=None, max_value=None, unit=None,
                 set_function=None, default_type=u"str", decimals=0, readonly=False, description=None,
                 set_value_when_set_function=True, tolerance=(0, 0), factor=1, offset=0, save_change=False,
                 allow_stream=True, **kwargs):
        dict.__init__(self)
        self._callbacks = {}
        self.attribute_id = attribute_id
        self.set_function = set_function
        self.default_type = default_type

        self[OFFSET] = offset
        self[FACTOR] = factor
        # Set attribute value flag when using set function.
        # This flag prevents change of attribute's value before value is actually set on device side.
        self.set_value_when_set_function = set_value_when_set_function
        if unit is None:
            unit = u""
        self[UNIT] = unit
        self.min_stream_timeout = kwargs.get("min_stream_timeout", 100) * 1e6
        default_type = kwargs.get(TYPE, default_type)
        if default_type == TYPE_LIST:
            self[TYPE] = TYPE_LIST
        elif isinstance(default_type, (list, collections.abc.KeysView)):
            self[TYPE] = TYPE_LIST
            self[TYPE_LIST_VALUES] = list(default_type)
        elif default_type == TYPE_ARRAY or isinstance(default_type, ndarray):
            self[TYPE] = TYPE_ARRAY
        else:
            default_type = dtype(default_type)
            self[TYPE] = dtype(default_type).name
            if default_type in self.integer_types:
                self[DECIMALS] = 0
                if min_value is None and default_type in self.unsigned_types:
                    self[MIN] = 0
                else:
                    self[MIN] = int(min_value) if min_value is not None else -math.inf
                self[MAX] = int(max_value) if max_value is not None else math.inf
                self.numerical = True
            elif default_type in self.float_types:
                self[DECIMALS] = decimals
                self[MIN] = float(min_value) if min_value is not None else -math.inf
                self[MAX] = float(max_value) if max_value is not None else math.inf
                self.numerical = True
        if self.numerical:
            self[TOLERANCE] = [float(tolerance[0]), float(tolerance[1])]
        self[READONLY] = readonly
        self[DESCRIPTION] = description
        self[VALUE] = default_value
        self[SETPOINT] = None
        self[PREVIOUS] = None
        self[SAVE_CHANGE] = save_change
        self[ALLOW_STREAM] = allow_stream
        self.update(kwargs)

    def _update_set_time(self):
        """
        Set last_update_time and last_update_time_diff for future reference.
        :return: None
        """
        update_time = perf_counter_ns()
        self.last_update_time_diff = update_time - self.last_update_time
        self.last_update_time = update_time

    def __setitem__(self, k, v):
        if k == VALUE:
            try:
                self[PREVIOUS] = self[VALUE]
            except KeyError:
                pass
            if isinstance(v, units.Quantity):
                try:
                    v = v.to(self[UNIT]).m
                except AttributeError:
                    v = v.m
        elif k == TOLERANCE:
            # Make sure that tolerance is float
            v = [float(t) for t in v]
        elif k == UNIT and v is not None:
            if v == u"%":
                v = u"percent"
            if v not in ("", None):
                if units.parse_unit_name(v) == ():
                    units.define(u"{}=1".format(v))
                    units.derived_units.append(v)
        try:
            if is_equal(self[k], v):
                return
        except KeyError:
            pass

        if k == VALUE and self.set_function and self.with_set_function:
            self[SETPOINT] = v
            setpoint = self.remove_offset_factor(v)
            # Release set function flag
            self.with_set_function = False
            if self.set_value_when_set_function:
                # Store actual attribute[k] value into dictionary
                super(Attribute, self).__setitem__(k, v)
            # Call set function
            if hasattr(self.set_function, "callback"):
                self.set_token = self.set_function(setpoint, callback=self.set_callback)
            else:
                # We don't have callback in a function, so just call it plain without callback
                self.set_token = self.set_function(setpoint)
            if not self.set_value_when_set_function:
                return
            self._update_set_time()
        else:
            # Store actual attribute[k] value into dictionary
            super(Attribute, self).__setitem__(k, v)
            self._update_set_time()

        self.handle_callbacks(k, v)

    def handle_callbacks(self, k, v):
        """
        When any parameter of attribute is set, call all relevant callbacks that are attached to the parameter.
        @param k: attribute key
        @type k: hashable
        @param v: attribute value
        @type v: object
        """
        # Call all relevant callbacks
        for callback, (max_update_rate, updated_at, key_filter) in self._callbacks.copy().items():
            # Check update rate, continue if it's less than max_update_rate
            if k == VALUE and max_update_rate is not None and (self.last_update_time - updated_at) <= max_update_rate:
                continue
            # Check if key_filter is set
            if key_filter is None:
                # Key can be any value, therefore callback with key and value
                callback(k, v)
            elif k == key_filter:
                # Since we have filter set, callback with just value
                callback(v)
            else:
                continue
            try:
                # Set last callback time
                self._callbacks[callback][1] = perf_counter_ns()
            except KeyError:
                # Callback no longer exists, skipping
                pass

    def attribute_copy(self):
        """
        Create copy of Attribute
        @return: Copy of this Attribute
        @rtype: Attribute
        """
        return Attribute(**self)

    @staticmethod
    def from_dict(dictionary):
        """
        Create Attribute from dictionary.
        @param dictionary: Dictionary
        @type dictionary: dict
        @return: Attribute if convertible
        @rtype: Attribute
        """
        if isinstance(dictionary, dict):
            if Attribute.is_attribute(dictionary):
                if dictionary[UNIT] not in ("", None):
                    try:
                        units.get_name(dictionary[UNIT])
                    except pint.errors.UndefinedUnitError:
                        units.define(u"{}=1".format(dictionary[UNIT]))
                        units.derived_units.append(dictionary[UNIT])
                return Attribute(**dictionary)
            else:
                for key, attribute in dictionary.items():
                    dictionary[key] = Attribute.from_dict(attribute)
            return dictionary
        else:
            return dictionary

    @staticmethod
    def is_attribute(dictionary):
        """
        Check if provided dictionary is convertible to Attribute.
        :param dictionary: dict
        :return: bool
        """
        try:
            return Attribute.attribute_keys.issubset(dictionary.keys())
        except (AttributeError, TypeError, ValueError):
            return False

    def attach_callback(self, callback, max_update_rate=None, key_filter=None):
        """
        Attach callback to Attribute.
        :param callback: callable function
        :param max_update_rate: max update rate in ms
        :return: None
        """
        assert callable(callback)

        if callback not in self._callbacks:
            if max_update_rate is not None:
                max_update_rate *= 1e6
            self._callbacks[callback] = [max_update_rate, perf_counter_ns(), key_filter]

    def detach_callback(self, callback):
        """
        Detach callback from attribute.
        :param callback: callable function
        :return: None
        """
        assert callable(callback)
        try:
            del self._callbacks[callback]
        except KeyError:
            pass  # Callback no longer exists

    def apply_offset_factor(self, value=None):
        """
        Apply offset and factor to attribute value.
        :param value: numerical value
        :return: mixed
        """
        if value is None:
            value = self[VALUE]

        if value is not None and self.numerical:
            return (value + self[OFFSET]) * self[FACTOR]
        elif self.default_type is bool:
            return self[FACTOR] == value
        else:
            return value

    def remove_offset_factor(self, value=None):
        """
        Remove offset and factor from attribute value.
        :param value: numerical value
        :return: mixed
        """
        if value is None:
            value = self[VALUE]

        if value is not None and self.numerical:
            return (value / self[FACTOR]) - self[OFFSET]
        elif self.default_type is bool:
            return self[FACTOR] == value
        else:
            return value

    def read_and_reset_token(self):
        """
        Return token and reset it back to 0.
        :return: int
        """
        token = self.set_token
        self.set_token = 0
        return token

    @staticmethod
    def list_attribute(attribute):
        """
        Create list from input attribute.
        Many methods relies on the fact that attribute is stored as list.
        :param attribute: mixed
        :return: list
        """
        if isinstance(attribute, list):
            return attribute
        elif isinstance(attribute, tuple):
            return list(attribute)
        else:
            return [attribute]

    def convert_units(self, value):
        """
        Convert input value to current attribute unit.
        :param value: units.Quantity
        :return: value or units.Quantity
        """
        assert isinstance(value, units.Quantity)
        if value.dimensionless:
            if self[UNIT] is None:
                return value
            else:
                return units.Quantity("{} {}".format(value.m, self[UNIT]))

        return value.to(self[UNIT])

    def within_limits(self, value):
        """
        Check if value is within limits.
        :param value:  units.Quantity
        :return: bool
        """
        assert isinstance(value, units.Quantity)
        return self.minimum() <= value <= self.maximum()

    def value(self):
        """
        Get attribute's value.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[VALUE], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[VALUE])
            except TypeError:
                return self[VALUE]
        else:
            return self[VALUE]

    def minimum(self):
        """
        Get attribute's minimum value.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[MIN], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[MIN])
            except TypeError:
                return self[MIN]
        else:
            return self[MIN]

    def maximum(self):
        """
        Get attribute's maximum value.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[MAX], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[MAX])
            except TypeError:
                return self[MAX]
        else:
            return self[MAX]

    def negative_tolerance(self):
        """
        Get attribute's negative tolerance in units.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[TOLERANCE][0], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[TOLERANCE][0])
            except TypeError:
                return self[TOLERANCE][0]
        else:
            return self[TOLERANCE][0]

    def positive_tolerance(self):
        """
        Get attribute's positive tolerance in units.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[TOLERANCE][1], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[TOLERANCE][1])
            except TypeError:
                return self[TOLERANCE][1]
        else:
            return self[TOLERANCE][1]

    def offset(self):
        """
        Get attribute's value offset in units.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[OFFSET], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[OFFSET])
            except TypeError:
                return self[OFFSET]
        else:
            return self[OFFSET]

    def setpoint(self):
        """
        Get attribute's setpoint in units.
        :return: Quantity
        """
        if self.numerical:
            try:
                return units.Quantity(self[SETPOINT], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[SETPOINT])
            except TypeError:
                return self[SETPOINT]
        else:
            return self[SETPOINT]

    def unit(self):
        """
        Get current unit.
        :return: str
        """
        return self[UNIT]



import functools
import gc
import hashlib
import operator
import os
import sys
from functools import reduce  # forward compatibility for Python 3

import numpy as np


def get_from_dict(data_dict, map_list):
    return reduce(operator.getitem, map_list, data_dict)


def set_in_dict(data_dict, map_list, value):
    get_from_dict(data_dict, map_list[:-1])[map_list[-1]] = value


sentinel = object()


def rgetattr(obj, attr, default=sentinel):
    if default is sentinel:
        _getattr = getattr
    else:
        def _getattr(obj, name):
            return getattr(obj, name, default)
    return functools.reduce(_getattr, [obj] + attr.split('.'))


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def get_next_file_index(directory, pattern):
    highest = -1
    for fname in os.listdir(directory):
        find = pattern.search(fname)
        if find is not None:
            if int(find.group(1)) > highest: highest = int(find.group(1))
    highest += 1
    return highest


def fullname(o):
    # o.__module__ + "." + o.__class__.__qualname__ is an example in
    # this context of H.L. Mencken's "neat, plausible, and wrong."
    # Python makes no guarantees as to whether the __module__ special
    # attribute is defined, so we take a more circumspect approach.
    # Alas, the module name is explicitly excluded from __qualname__
    # in Python 3.
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__  # Avoid reporting __builtin__
    else:
        return module + '.' + o.__class__.__name__


def get_object_refs(o):
    ref_count = sys.getrefcount(o)
    if ref_count > 0:
        print("Number of referrers of object {} is {}".format(o, ref_count))
        for index, referrer in enumerate(gc.get_referrers(o)):
            print(index, referrer)


def is_equal(a, b):
    c = a == b
    if isinstance(c, np.ndarray):
        c = c.all()
    return c


def compute_MD5(my_string):
    m = hashlib.md5()
    m.update(my_string.encode('utf-8'))
    return m.hexdigest()



import asynchat
import socket
from threading import RLock, Lock

from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsControlLoops import control_asyncore_loop


class DeviceSocket(asynchat.async_chat, Device):
    terminator = b"\r\n"
    ac_in_buffer_size = 4096
    ac_out_buffer_size = 4096
    host = None
    port = None

    def __init__(self, host, port, device_id=None, config=None):
        Device.__init__(self, device_id, config)
        self.asyncore_loop = control_asyncore_loop
        self.host = host
        self.port = port
        self.send_lock = RLock()
        asynchat.async_chat.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setblocking(False)
        self.buffer = []
        self.connect((host, port))

    def connect(self, address):
        try:
            asynchat.async_chat.connect(self, address)
            self.device_connection_poller.add_connecting_device(self)
        except socket.error:
            u"""
            Socket could not be connected. Non fatal error.
            We don't want to reconnect immediately but rather wait for connection timeout.
            Set devices as connecting, add it to device poller and wait until connection timeout.  
            """
            self.logger.exception("Socket creation exception")
            self.connecting = True
            self.device_connection_poller.add_connecting_device(self)
        except (OverflowError, TypeError, OSError):
            self.logger.exception("Socket creation exception")
            self.close()

    def initiate_send(self):
        try:
            with self.send_lock:
                asynchat.async_chat.initiate_send(self)
        except IndexError:
            # Sending queue is already empty
            # Ignore this error
            self.logger.error(u"Send buffer is already empty")
            pass

    def collect_incoming_data(self, data):
        self.buffer.append(data.decode("utf-8"))

    def found_terminator(self):
        self.handle_readout(self.buffer)
        self.buffer = []

    def handle_connect(self):
        Device.handle_connect(self)

    def handle_connect_event(self):
        try:
            asynchat.async_chat.handle_connect_event(self)
            self.handle_configuration_event()
        except socket.error:
            self.logger.exception(u"Socket creation exception")
            self.close()

    def close_connection(self):
        self.stop_polling()
        # return asynchat.async_chat.close()
        return self.close()

    def close(self):
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        try:
            asynchat.async_chat.close(self)
        except OSError:
            pass
        self.set_status(STATUS_DISCONNECTED)

    def handle_expt_event(self):
        self.logger.error("Socket error")
        self.connection_error = True
        asynchat.async_chat.handle_expt_event(self)

    def handle_error(self):
        self.logger.exception("Uncaught exception")
        self.handle_response_error()

    def handle_configuration(self):
        raise NotImplementedError(u"Must be implemented in subclass")



from PyQt5.QtCore import pyqtSignal, pyqtSlot, QSize
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QWidget, QPushButton

from kamzik3.constants import *
from kamzik3.devices.observer import Observer
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.templates.deviceGroupTemplate import Ui_Form
from kamzik3.snippets.snippetsYaml import YamlSerializable


class DeviceGroupWidget(Ui_Form, QWidget, Observer, YamlSerializable):
    sig_set_status_color = pyqtSignal("QString")
    model_image = None
    signal_detach_request = pyqtSignal("QString")
    signal_attach_request = pyqtSignal("QString")
    attached = True

    def __init__(self, title, model_image=None, config=None, parent=None):
        QWidget.__init__(self, parent)
        self.devices = []
        self.config = config
        if model_image is not None:
            self.model_image = QIcon(model_image)
        self.title = title
        self.setupUi(self)
        self.sig_set_status_color.connect(self.set_status_color)
        self.button_group_name.setText("{}  ".format(title))

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    def setupUi(self, Form):
        super(DeviceGroupWidget, self).setupUi(Form)
        if self.model_image is not None:
            self.button_device_group.setIcon(self.model_image)
        self.button_device_group.resizeEvent = self.button_resized_event

    def add_device(self, device):
        """
        Add new device into group
        :param device:
        :return:
        """
        self.devices.append(device)
        self.attach_to_subject(device)

    def subject_update(self, key, value, subject):
        """
        Check all statuses if only one status changed.
        This way we ensure that group status represents all device statuses.
        :param key:
        :param value:
        :param subject:
        :return:
        """
        if key == ATTR_STATUS:
            status_priorities = []
            for device in self.devices:
                priority = STATUSES_PRIORITY.index(device[ATTR_STATUS][VALUE])
                status_priorities.append(priority)

            group_status = STATUSES_PRIORITY[min(status_priorities)]
            status_color = DeviceWidget.status_colors.get(group_status, "blue")
            self.sig_set_status_color.emit(status_color)

    @pyqtSlot("QString")
    def set_status_color(self, color):
        self.frame_label_area.setStyleSheet("QFrame {{background-color:{}}}".format(color))

    def close(self):
        for device in self.devices:
            self.detach_subject(device)

    @pyqtSlot("QIcon")
    def slot_show_model(self, model_image):
        self.button_device_group.setIcon(model_image)

    @pyqtSlot("QIcon")
    def slot_hide_model(self, model_image):
        self.button_device_group.setIcon(self.model_image)

    def button_resized_event(self, QResizeEvent):
        """
        This method resize group to max of 200 pixels width
        :param QResizeEvent:
        :return:
        """
        w, h = QResizeEvent.size().width(), QResizeEvent.size().height()
        if w >= 200:
            h = w = 200
        else:
            h = w
        self.button_device_group.setIconSize(QSize(w, h))
        return QPushButton.resizeEvent(self.button_device_group, QResizeEvent)

    def toggle_attach(self):
        """
        Toggle tab attachment
        :return:
        """
        self.attached = not self.attached
        if self.attached:
            self.button_toggle_attach.setIcon(QIcon(":/icons/icons/detach.png"))
            self.signal_attach_request.emit(self.title)
        else:
            self.button_toggle_attach.setIcon(QIcon(":/icons/icons/attach.png"))
            self.signal_detach_request.emit(self.title)



import struct
import time

import numpy
import pandas

import kamzik3
from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceInficonQcm(DeviceSocket):
    terminator = None
    push_commands_max = 1
    command_encoding = None
    device_packet_error = {
        'C': "invalid checksum",
        'F': "illegal format",
        'I': "invalid message",
        'M': "too many commands",
        'O': "response length is larger than"
    }
    device_response_error = {
        'A': "illegal command",
        'B': "illegal parameter value",
        'D': "illegal ID",
        'E': "data not available",
        'F': "cannot do now",
        'L': "length error",
        'P': "prior command failed"
    }

    def __init__(self, host, port=2101, device_id=None, config=None):
        self._materials = pandas.read_csv("./resources/inficon_qcm_materials.txt", sep="\t", names=["Formula", "Name"])
        self.materials = (self._materials["Formula"] + " - " + self._materials["Name"]).to_list()
        DeviceSocket.__init__(self, host, port, device_id, config)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_PRODUCT_TYPE, readonly=True)
        self.create_attribute(ATTR_SOFTWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_FIRMWARE, readonly=True)
        self.create_attribute(ATTR_SENSORS_COUNT, default_value=8, readonly=True)
        self.create_attribute(ATTR_ACTIVE_PROCESS, default_value=1, default_type=int, readonly=True)

        for i in range(1, self.get_value(ATTR_SENSORS_COUNT) + 1):
            self.create_attribute(f"{ATTR_MATERIAL}_{i}", default_value=None, default_type=self.materials,
                                  set_function=lambda v, i=i: self.set_material(i, v), group=GROUP_MATERIALS)
            self.create_attribute(f"{ATTR_OUTPUT}_{i}", default_value=None,
                                  set_function=lambda v, i=i: self.set_output_name(i, v), group=GROUP_OUTPUTS)
            self.create_attribute(f"{ATTR_PROCESS}_{i}", default_value=None,
                                  set_function=lambda v, i=i: self.set_process_name(i, v), group=GROUP_PROCESSES)
            self.create_attribute(f"{ATTR_USER_MESSAGE}_{i}", default_value=None,
                                  set_function=lambda v, i=i: self.set_user_message(i, v), group=GROUP_USER_MESSAGES)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.get_revision()
        self.get_version()
        for i in range(1, self.get_value(ATTR_SENSORS_COUNT) + 1):
            self.send_command_data("QN", self.t_byte(i))
            # Disable multipoint for each sensor
            self.send_command_data("UM", self.t_byte(48) + self.t_byte(i) + self.t_int(0))
            self.send_command_data("QO", self.t_byte(i))
            self.send_command_data("QK", self.t_byte(i))
            self.send_command_data("QV", self.t_byte(i))
            # Set failure action to STOP = 2
            self.send_command_data("UM", self.t_byte(115) + self.t_byte(i) + self.t_int(2))
            # Disable backup sensor
            self.send_command_data("UM", self.t_byte(119) + self.t_byte(i) + self.t_int(0))
            # Set tooling to 100%
            self.send_command_data("UM", self.t_byte(120) + self.t_byte(i) + self.t_float(100))
            # Set quality control to 0
            self.send_command_data("UM", self.t_byte(121) + self.t_byte(i) + self.t_int(0))
            # Set quality counts to 0
            self.send_command_data("UM", self.t_byte(122) + self.t_byte(i) + self.t_int(0))
            # Set stability single to 0
            self.send_command_data("UM", self.t_byte(123) + self.t_byte(i) + self.t_int(0))
            # Set stability total to 0
            self.send_command_data("UM", self.t_byte(124) + self.t_byte(i) + self.t_int(0))
            # Set output type to NO
            self.send_command_data("UT", self.t_byte(i) + self.t_byte(0))
            # Set calibration to uncalibrated = 999.9
            for j in range(90, 97, 1):
                self.send_command_data("UM", self.t_byte(j) + self.t_byte(i) + self.t_float(999.9))

        self.get_firmware(callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DeviceSocket.handle_readout(self, readout_buffer)
        error_flag = bool(output[0])
        timer_tick = output[1]
        ack = output[2]
        if error_flag:
            raise DeviceError(self.device_packet_error.get(chr(ack)))
        if ack != 6:
            print(command)
            raise DeviceError(self.device_response_error.get(chr(ack)))

        first_ch = chr(command[2])
        if first_ch == "H":
            command_group, command_id = first_ch, command[3]
            command_data = output[3:]
        elif first_ch in ("Q", "R", "U", "S"):
            command_group, command_id = chr(command[2]) + chr(command[3]), command[4]
            command_data = output[3:]

        if command_group == "H":
            if command_id == 1:
                self.set_value(ATTR_PRODUCT_TYPE, self.f_string(command_data))
            elif command_id == 2:
                version_number = self.f_int(command_data[:4])
                compatibility_version = self.f_int(command_data[4:8])
                range_version = self.f_int(command_data[8:])
                self.set_value(ATTR_SOFTWARE_VERSION,
                               f"VN:{version_number}, CV:{compatibility_version}, RV:{range_version}")
            elif command_id == 3:
                firmware_version = self.f_float(command_data)
                self.set_value(ATTR_FIRMWARE, f"{firmware_version:.3f}")
        elif command_group == "QN":
            material_index = command[4]
            formula = self.f_string(command_data)
            if formula != "" and " - " in formula:
                formula, library_index = formula.split(" - ")
                material = self.materials[int(library_index)]
                self.set_raw_value([GROUP_MATERIALS, f"{ATTR_MATERIAL}_{material_index}"], material)
        elif command_group == "SG":
            for channel in range(self.get_value(ATTR_SENSORS_COUNT)):
                self.notify((channel, ATTR_STATE), int(command_data[channel]))
        elif command_group == "QO":
            output_index = command[4]
            name = self.f_string(command_data)
            self.set_raw_value([GROUP_OUTPUTS, f"{ATTR_OUTPUT}_{output_index}"], name)
        elif command_group == "QK":
            output_index = command[4]
            name = self.f_string(command_data)
            self.set_raw_value([GROUP_PROCESSES, f"{ATTR_PROCESS}_{output_index}"], name)
        elif command_group == "QS" and int(command[4]) == 2:
            channel = int(command[5]) - 1
            self.notify((channel, ATTR_SENSOR_TYPE), self.f_int(command_data))
        elif command_group == "QS" and int(command[4]) == 4:
            channel = int(command[5]) - 1
            self.notify((channel, ATTR_AUTO_Z), bool(self.f_int(command_data)))
        elif command_group == "QV":
            output_index = command[4]
            value = self.f_string(command_data)
            self.set_raw_value([GROUP_USER_MESSAGES, f"{ATTR_USER_MESSAGE}_{output_index}"], value)
        elif command_group == "SS":
            channel = int(command[5]) - 1
            if channel == -1:
                if command_id == 6:
                    for channel, bit_index in enumerate(range(0, len(command_data), 4)):
                        self.notify((channel, ATTR_Z_RATIO), self.f_float(command_data[bit_index:bit_index + 4]))
                elif command_id == 4:
                    for channel, bit_index in enumerate(range(0, len(command_data), 8)):
                        self.notify((channel, ATTR_FUNDAMENTAL_FREQUENCY),
                                    self.f_double_int(command_data[bit_index:bit_index + 8]))
                elif command_id == 7:
                    for channel, bit_index in enumerate(range(0, len(command_data), 8)):
                        self.notify((channel, ATTR_AHARMONIC_FREQUENCY),
                                    self.f_double_int(command_data[bit_index:bit_index + 8]))
                elif command_id == 12:
                    for channel, bit_index in enumerate(range(0, len(command_data), 4)):
                        self.notify((channel, ATTR_SENSOR_THICKNESS),
                                    self.f_float(command_data[bit_index:bit_index + 4]))
                elif command_id == 0:
                    for channel, bit_index in enumerate(range(0, len(command_data), 1)):
                        self.notify((channel, ATTR_CRYSTAL_LIFE),
                                    int.from_bytes(command_data[bit_index:bit_index + 1], "little"))
                elif command_id == 13:
                    for channel, bit_index in enumerate(range(0, len(command_data), 4)):
                        self.notify((channel, ATTR_RAW_RATE),
                                    self.f_float(command_data[bit_index:bit_index + 4]))
            else:
                if command_id == 6:
                    self.notify((channel, ATTR_Z_RATIO), self.f_float(command_data))
                elif command_id == 4:
                    self.notify((channel, ATTR_FUNDAMENTAL_FREQUENCY), self.f_double_int(command_data))
                elif command_id == 7:
                    self.notify((channel, ATTR_AHARMONIC_FREQUENCY), self.f_double_int(command_data))
                elif command_id == 12:
                    self.notify((channel, ATTR_SENSOR_THICKNESS), self.f_float(command_data))
                elif command_id == 0:
                    self.notify((channel, ATTR_CRYSTAL_LIFE), int.from_bytes(command_data, "little"))
                elif command_id == 14:
                    self.notify((channel, ATTR_RAW_RATE), self.f_float(command_data))
        elif command_group == "SL":
            self.notify((self.get_value(ATTR_ACTIVE_PROCESS)-1, ATTR_LAYER_THICKNESS), self.f_float(command_data))

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, InficonQcmSensorChannel):
                    kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)

    def notify_channels(self, attribute, value):
        for channel in range(self.get_value(ATTR_SENSORS_COUNT)):
            self.notify((channel, attribute), value)

    def set_material(self, material_index, material):
        library_index = self.materials.index(material)
        # Set material from library into controller material index 1-32
        command_data = self.t_byte(38) + self.t_byte(material_index) + self.t_byte(library_index)
        self.send_command_data("RG", command_data)
        # Rename material, add library index
        formula, full_name = material.split(" - ")
        controller_name = f"{formula} - {library_index}"
        command_data = self.t_byte(material_index) + self.t_str(controller_name)
        self.send_command_data("UN", command_data)

    def set_output_name(self, output_index, name):
        command_data = self.t_byte(output_index) + self.t_str(name)
        self.send_command_data("UO", command_data)

    @staticmethod
    def f_string(readout):
        return readout[:-1].decode()

    @staticmethod
    def f_int(readout):
        return int.from_bytes(readout, "little")

    @staticmethod
    def f_float(readout):
        return struct.unpack('f', readout)[0]

    @staticmethod
    def f_double_int(readout):
        return struct.unpack('q', readout)[0]

    @staticmethod
    def t_int(int_value):
        return int(int_value).to_bytes(4, "little")

    @staticmethod
    def t_float(float_value):
        return bytearray(struct.pack("f", float_value))

    @staticmethod
    def t_byte(int_value):
        return int(int_value).to_bytes(1, "little")

    @staticmethod
    def t_str(str_value):
        return str_value.encode() + b"\0"

    def start_polling(self):
        DeviceSocket.start_polling(self)
        self.poll_command(self.send_command_data("SG", self.t_byte(10), execute=False), 1000)
        self.poll_command(self.send_command_data("SS", self.t_byte(4) + self.t_byte(0), execute=False), 1000)
        self.poll_command(self.send_command_data("SS", self.t_byte(6) + self.t_byte(0), execute=False), 1000)
        self.poll_command(self.send_command_data("SS", self.t_byte(7) + self.t_byte(0), execute=False), 1000)
        self.poll_command(self.send_command_data("SS", self.t_byte(12) + self.t_byte(0), execute=False), 1000)
        self.poll_command(self.send_command_data("SS", self.t_byte(0) + self.t_byte(0), execute=False), 1000)
        self.poll_command(self.send_command_data("SS", self.t_byte(13) + self.t_byte(0), execute=False), 1000)
        self.poll_command(self.send_command_data("SL", self.t_byte(4) + self.t_byte(1), execute=False), 1000)

    def get_available_materials(self):
        materials_list = list()
        for i in range(1, 33):
            materials_list.append(self.get_value([GROUP_MATERIALS, f"{ATTR_MATERIAL}_{i}"]))
        return materials_list

    def send_command_data(self, command_group, command_data=b"", callback=None, with_token=False,
                          returning=True, execute=True):
        message = command_group.encode() + command_data
        message_length = len(message).to_bytes(2, "little")

        # In order to have final command, checksum needs to be calculated
        checksum = sum(message) % 256
        packet = message_length + message + checksum.to_bytes(1, "little")
        if execute:
            self.command(packet, callback, with_token, returning)
        else:
            return packet

    def collect_incoming_data(self, data):
        """
        This method serves to collect data from communication port.
        After device replies, we have to confirm receiving it by sending x06.
        Accumulate data until we have expected response length and its checksum is correct.
        :param data: input data stream
        :return:
        """
        # Append new data into buffer
        self.buffer.append(data)
        # Generate bytearray
        full_response = b"".join(self.buffer)
        response_length = int.from_bytes(full_response[:2], "little")
        if len(full_response) >= response_length + 3:
            response = full_response[:response_length + 3]
            self.buffer = [full_response[response_length:]] if response_length > len(full_response) else []
            response_checksum = sum(full_response[2:-1]) % 256
            control_checksum = response[-1]

            response = response[2:-1]
            if response_checksum == control_checksum:
                self.handle_readout(response)
            else:
                # checksum is not OK, request packet again
                self.buffer = []
                # Ask for response again
                DeviceError("Checksum error!")

    def get_revision(self):
        self.send_command_data("H", self.t_byte(1))

    def get_version(self):
        self.send_command_data("H", self.t_byte(2))

    def get_firmware(self, callback=None):
        self.send_command_data("H", self.t_byte(3), callback=callback)

    def set_output_name(self, output, name):
        command_data = self.t_byte(output) + self.t_str(name)
        self.send_command_data("UO", command_data)

    def set_process_name(self, process, name):
        command_data = self.t_byte(process) + self.t_str(name)
        self.send_command_data("UK", command_data)

    def set_user_message(self, message, value):
        command_data = self.t_byte(message) + self.t_str(value)
        self.send_command_data("UV", command_data)

    @expose_method()
    def init_measurement(self):
        self.set_run_number(1)
        self.reset_thickness(1)
        self.set_start_layer(1)

    @expose_method()
    def stop_measurement(self):
        pass

    def set_active_process(self, process):
        command_data = self.t_byte(1) + self.t_int(process)
        self.send_command_data("UG", command_data)
        self.set_raw_value(ATTR_ACTIVE_PROCESS, process)

    def set_start_layer(self, layer):
        command_data = self.t_byte(2) + self.t_int(layer)
        self.send_command_data("UG", command_data)

    def set_run_number(self, value):
        command_data = self.t_byte(3) + self.t_int(value)
        self.send_command_data("UG", command_data)

    def remove_layers(self, process):
        command_data = self.t_byte(28) + self.t_byte(process)
        self.send_command_data("RG", command_data)

    def reset_thickness(self, layer):
        command_data = self.t_byte(1) + self.t_byte(layer)
        self.send_command_data("RL", command_data)

    def get_layer_thickness(self, layer):
        command_data = self.t_byte(4) + self.t_byte(layer)
        self.send_command_data("SL", command_data)

class InficonQcmSensorChannel(DeviceChannel):
    sensor_state = [
        "Good Crystal and active",
        "Failed Crystal and active",
        "Invalid Measurement on Crystal and active",
        "Good Crystal and inactive",
        "Failed Crystal and inactive",
        "Invalid Measurement on Crystal and inactive"
    ]
    sensor_type = [
        "Single", "XtalTwo", "XtalSix", "Xtal12", "Generic"
    ]

    def __init__(self, device, channel, device_id=None, config=None):
        self.sensor_id = DeviceInficonQcm.t_int(channel + 1)
        DeviceChannel.__init__(self, device, channel, device_id, config)

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)
        self.create_attribute(ATTR_STATE, readonly=True)
        self.add_attribute(ATTR_MATERIAL,
                           self.device.get_attribute([GROUP_MATERIALS, f"{ATTR_MATERIAL}_{self.channel + 1}"]))
        self.create_attribute(ATTR_SENSOR_TYPE, default_type=self.sensor_type, set_function=self.set_sensor_type)
        self.create_attribute(ATTR_AUTO_Z, default_type=bool, set_function=self.set_auto_z)
        self.create_attribute(ATTR_Z_RATIO, default_value=None, default_type=float, readonly=True, decimals=9)
        self.create_attribute(ATTR_FUNDAMENTAL_FREQUENCY, unit="Hz", default_value=None, default_type=numpy.double,
                              readonly=True, decimals=9, factor=0.000873114913702011)
        self.create_attribute(ATTR_AHARMONIC_FREQUENCY, unit="Hz", default_value=None, default_type=numpy.double,
                              readonly=True, decimals=9, factor=0.000873114913702011)
        self.create_attribute(ATTR_SENSOR_THICKNESS, unit="nm", default_value=None, default_type=float,
                              readonly=True, decimals=9, factor=0.1)
        self.create_attribute(ATTR_CRYSTAL_LIFE, unit="%", default_value=None, default_type=int,
                              readonly=True)
        self.create_attribute(ATTR_RAW_RATE, unit="nm/s", default_value=None, default_type=float,
                              readonly=True, decimals=9, factor=0.1)
        self.create_attribute(ATTR_SHUTTER_STROKES, default_value=0, default_type=int,
                              readonly=True, save_change=True, min_value=0)
        self.create_attribute(ATTR_SHUTTER_OPEN, default_value=False, default_type=bool,
                              readonly=True)
        self.create_attribute(ATTR_LAYER_THICKNESS, unit="nm", default_value=None, default_type=float,
                              readonly=True, decimals=9, factor=0.1)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.close_shutter()
            self.configured = True
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.get_sensor_type()
        self.get_auto_z()
        self.set_shutter_logic()
        self.set_channel_to_material(callback=_finish_configuration)
        # Assign material # to process #
        self.device.send_command_data("UP", self.device.t_byte(1) + self.device.t_byte(self.channel+1) + self.device.t_byte(self.channel+1) + self.sensor_id)
        self.get_z_ratio()
        self.get_fundamental_frequency()
        self.get_aharmonic_frequency()
        self.get_sensor_thickness()
        self.get_raw_rate()
        _finish_configuration()

    def get_sensor_type(self):
        command_data = DeviceInficonQcm.t_byte(2) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("QS", command_data)

    def get_auto_z(self):
        command_data = DeviceInficonQcm.t_byte(4) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("QS", command_data)

    def get_z_ratio(self):
        command_data = DeviceInficonQcm.t_byte(6) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("SS", command_data)

    def get_fundamental_frequency(self):
        command_data = DeviceInficonQcm.t_byte(4) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("SS", command_data)

    def get_aharmonic_frequency(self):
        command_data = DeviceInficonQcm.t_byte(7) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("SS", command_data)

    def get_sensor_thickness(self):
        command_data = DeviceInficonQcm.t_byte(12) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("SS", command_data)

    def get_crystal_life(self):
        command_data = DeviceInficonQcm.t_byte(0) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("SS", command_data)

    def get_raw_rate(self):
        command_data = DeviceInficonQcm.t_byte(13) + DeviceInficonQcm.t_byte(self.channel + 1)
        self.device.send_command_data("SS", command_data)

    def set_sensor_type(self, type):
        type_index = self.sensor_type.index(type)
        command_data = DeviceInficonQcm.t_byte(2) + DeviceInficonQcm.t_byte(self.channel + 1) + DeviceInficonQcm.t_int(
            type_index)
        self.device.send_command_data("US", command_data)

    def set_auto_z(self, value):
        command_data = DeviceInficonQcm.t_byte(4) + DeviceInficonQcm.t_byte(self.channel + 1) + DeviceInficonQcm.t_int(
            value)
        self.device.send_command_data("US", command_data)

    def subject_update(self, key, value, subject):
        DeviceChannel.subject_update(self, key, value, subject)
        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            if key[1] == ATTR_STATE:
                self.set_raw_value(key[1], self.sensor_state[value])
            elif key[1] == ATTR_SENSOR_TYPE:
                self.set_raw_value(key[1], self.sensor_type[value])
            else:
                self.set_raw_value(key[1], value)

    @expose_method()
    def open_shutter(self):
        index = 2 * (self.channel + 1)
        command_data = DeviceInficonQcm.t_byte(6) + DeviceInficonQcm.t_byte(index)
        self.device.send_command_data("RG", command_data)
        command_data = DeviceInficonQcm.t_byte(5) + DeviceInficonQcm.t_byte(index - 1)
        self.device.send_command_data("RG", command_data)
        if not self.get_value(ATTR_SHUTTER_OPEN):
            self.set_value(ATTR_SHUTTER_STROKES, self.get_value(ATTR_SHUTTER_STROKES) + 1)
            self.set_value(ATTR_SHUTTER_OPEN, True)

    @expose_method()
    def close_shutter(self):
        index = 2 * (self.channel + 1)
        command_data = DeviceInficonQcm.t_byte(6) + DeviceInficonQcm.t_byte(index - 1)
        self.device.send_command_data("RG", command_data)
        command_data = DeviceInficonQcm.t_byte(5) + DeviceInficonQcm.t_byte(index)
        self.device.send_command_data("RG", command_data)
        if self.get_value(ATTR_SHUTTER_OPEN):
            self.set_value(ATTR_SHUTTER_STROKES, self.get_value(ATTR_SHUTTER_STROKES) + 1)
            self.set_value(ATTR_SHUTTER_OPEN, False)

    @expose_method()
    def measure_layer_thickness(self):
        self.device.set_active_process(self.channel + 1)
        self.device.get_layer_thickness(1)

    @expose_method()
    def clear_shutter_strokes(self):
        self.set_value(ATTR_SHUTTER_STROKES, 0)

    def set_channel_to_material(self, callback=None):
        command_data = DeviceInficonQcm.t_byte(114) + DeviceInficonQcm.t_byte(self.channel + 1) + self.sensor_id
        self.device.send_command_data("UM", command_data, callback=callback)

    def set_shutter_logic(self):
        # Count starting index for Logic commands
        index = 2 * (self.channel + 1)
        # Setting logic for Shutter Open
        self.device.set_value([GROUP_USER_MESSAGES, f"{ATTR_USER_MESSAGE}_{index}"],
                              f"Shutter {self.channel + 1} open")
        logic_statements = [DeviceInficonQcm.t_byte(0x7C),
                            DeviceInficonQcm.t_byte(0x20), DeviceInficonQcm.t_byte(0x41),
                            DeviceInficonQcm.t_byte(self.channel + 1), DeviceInficonQcm.t_byte(0x65),
                            DeviceInficonQcm.t_byte(index - 1), DeviceInficonQcm.t_byte(0x64),
                            DeviceInficonQcm.t_byte(index), DeviceInficonQcm.t_byte(0x03)]
        command_data = [DeviceInficonQcm.t_byte(index - 1),
                        DeviceInficonQcm.t_byte(len(logic_statements))] + logic_statements
        self.device.send_command_data("UL", b"".join(command_data))
        # Setting logic for Shutter Close
        self.device.set_value([GROUP_USER_MESSAGES, f"{ATTR_USER_MESSAGE}_{index - 1}"],
                              f"Shutter {self.channel + 1} closed")
        logic_statements = [DeviceInficonQcm.t_byte(0x7C),
                            DeviceInficonQcm.t_byte(0x20), DeviceInficonQcm.t_byte(0x42),
                            DeviceInficonQcm.t_byte(self.channel + 1), DeviceInficonQcm.t_byte(0x65),
                            DeviceInficonQcm.t_byte(index), DeviceInficonQcm.t_byte(0x64),
                            DeviceInficonQcm.t_byte(index - 1), DeviceInficonQcm.t_byte(0x03)]
        command_data = [DeviceInficonQcm.t_byte(index),
                        DeviceInficonQcm.t_byte(len(logic_statements))] + logic_statements
        self.device.send_command_data("UL", b"".join(command_data))
